#!/usr/bin/python3
#  coding: utf-8
#
#       coupling_OF_fistr.py
#
#           指定したpatchの
#               faceCenter座標
#               faceを構成するnodeNo（順番）
#               point座標
#           を取得し、OpenFOAMとFrontISTRを連成させる。
#
#   23/01/26    新規作成
#      03/17    getSavedFolders:毎回連成する時、facePressデータが削除され
#               latestTime以外の再計算が実現できなかった事を修正。
#      03/18    getNextCouplingTime:nStepOF=2,writeIntervalの次のstepで
#               hangupする事がある為、修正。
#      05/14    runCoupling:chtss計算を判断する為の変数skipSTSを追加
#               getNextCouplingTime:chtss計算を判断する為の変数skipStepsTempSを追加
#               temp計算後の次のstepで熱歪を計算させるため。
#      05/17    runCoupling,getNextCouplingTime:chtssのcouplingの方法を修正
#      06/13    OpenFOAMとFronISTRを並列で実行させる様に大幅修正。
#      06/26    連成解析する時、fsi、chtssで各々変位を合算するが、OpenFOAM側では
#               fsiとchtssが入り乱れ、読み込む変位のtimeIndexがずれる事がある。
#               fsi,chtssの場合は、送信dataのtitle内に、step数を含める様に変更。
#      07/01    deltaTを変更して、restartした時、step数が狂って、Hangupする事を修正
#      07/25    createInitialTemperatureFile:mshFileを直接文字列で指定したものを
#               datファイル内の記述名で指定（バグ）
#

import sys, os
import glob
import shutil
import subprocess
import time
import mmap
import multiprocessing
import math
import createPointCloudFromOpenFoam as createPtc
import pyTreeFoam_ptc as pyTreeFoam
import pyFistr_ptc as pyFistr
import mappingFromPointCloud as mapping
import pyCoupling

OF_pid = ""             #OpenFOAMのpid
waitCount = 5000        #最長で50sまで待つ
waitTime = 0.01
delTimes = []           #不要な時間のlist
delSteps = []           #不要なstepのlist
stepTimeDict = {}       #writeTimeとstepの対応list

commonSize = 10000


#
#  putProcStream
#----------------
def putProcStream(procNo, title, text):
    """ 指定したprocNoのstreamに文字をセットする"""
    size = commonSize
    myPid = os.getpid()
    header = "PID:" + str(myPid) + " title:" + title
    cont = header + "\01" + text + "\04"
    cont = cont.encode()
    dataList = []
    for i in range(0, len(cont), size):
        data = cont[i:i+size]
        dataList.append(data)
    #textをstreamにセットする
    for i in range(len(dataList)):
        if i == 0:
            waitUntilBufferInitialSet(procNo)
        else:
            waitUntilBufferEmpty(procNo)
        data = dataList[i]
        mmCommFile = mmCommFiles[int(procNo)]
        mmCommFile.seek(0)
        mmCommFile.write(data)
        mmCommFile.flush()
    mmCommFile.seek(commonSize-1)
    mmCommFile.write(b"\04")        #最後に書き込み
    mmCommFile.flush()
    waitUntilBufferEmpty(procNo)    #相手が読むまで待つ

#
#  getProcSteam
#---------------
def getProcStream(procNo, title):
    """ 指定したprocNoのstraemからtitleと文字列を取り出す"""
    waitUntilBufferSet(procNo, title)
    mmCommFile = mmCommFiles[int(procNo)]
    #buffer読み込み
    mmCommFile.seek(0)
    conts = mmCommFile.read().decode()
    #bufferクリア
    mmCommFile.seek(0)
    mmCommFile.write(b"\00\00")
    mmCommFile.seek(commonSize-1)
    mmCommFile.write(b"\00")
    mmCommFile.flush()
    #streamのheaderを取得
    ns = conts.find("\01")
    header = conts[:ns]
    #header内のtitleを取得
    nt = header.find("title:")
    title = header[nt + len("title:"):]
    #streamのtextを取得
    text = ""
    cont = conts[ns+1:]
    ne = cont.find("\04")
    if ne >= 0:
        text = cont[:ne]
        return title, text
    text += cont
    while True:
        waitUntilBufferSet(procNo, title)
        #読み込み
        mmCommFile.seek(0)
        cont = mmCommFile.read().decode()
        #bufferクリア
        mmCommFile.seek(0)
        mmCommFile.write(b"\00\00")
        mmCommFile.seek(commonSize-1)
        mmCommFile.write(b"\00")
        mmCommFile.flush()
        #読み込み内容を処理
        ne = cont.find("\04")
        if ne >= 0:
            text += cont[:ne]
            break
        text += cont
    return title, text

#
#  getProcStreamLines
#---------------------
def getProcStreamLines(procNo, title):
    """ 指定したprocNoのstraemからlinesを取り出す"""
    _title, text = getProcStream(procNo, title)
    if text[-1] == "\n":
        text = text[:-1]
    lines = text.split("\n")
    lines = list(map(lambda x: x+"\n", lines))
    return lines

#
#  waitUntilBufferInitialSet
def waitUntilBufferInitialSet(procNo):
    """ bufferがemptyの場合、直ぐにPIDをセットし、bufferを使用中にする"""
    mmCommFile = mmCommFiles[int(procNo)]
    count = waitCount
    sleepTime = waitTime
    bPid = str(os.getpid()).encode()
    while True:
        mmCommFile.flush()
        if mmCommFile[:2] == b"\00\00" and mmCommFile[-1:] == b"\00":
            mmCommFile.write(bPid)
            mmCommFile.flush()
            mmCommFile.seek(0)
            break
        count -= -1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  waitUntilBufferEmpty
def waitUntilBufferEmpty(procNo):
    """ 送信可能になるまで待つ。bufferが空になるまで"""
    mmCommFile = mmCommFiles[int(procNo)]
    count = waitCount
    sleepTime = waitTime
    while True:
        if mmCommFile[:2] == b"\00\00" and mmCommFile[-1:] == b"\00":
            break
        count -= -1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  waitUntilBufferSet
def waitUntilBufferSet(procNo, title):
    """ streamに文字がセットされるまで待つ"""
    mmCommFile = mmCommFiles[int(procNo)]
    count = waitCount
    sleepTime = waitTime
    while True:
        mmCommFile.flush()
        #最初と最後がb"\00"以外？
        if mmCommFile[:2] != b"\00\00" and mmCommFile[-1:] != b"\00":
            break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return  

#-------------------
#  clearDataFolder
#-------------------
def clearDataFolder():
    """ dataFolder内をクリアする。"""
    #couplingのdatafolder内をクリア
    coupDir = fluidCaseDir + "/coupling_FrontISTR"
    files = glob.glob(coupDir + "/data/*")
    for file in files:
        name = os.path.basename(file)
        if name != "times":
            os.remove(file)
    #FrontISTRのlogファイルを初期化
    name = fluidCaseDir + "/solidSolve.log"
    f = open(name, "w"); f.close()

#
#  clearTempFilesInDataFolder
#-----------------------------
def clearTempFilesInDataFolder():
    """ dataFolder内のtemporaryファイル（tmファイル）を削除する"""
    startTime = float(couplingDict["startTime"])
    coupDir = fluidCaseDir + "/coupling_FrontISTR"
    files = glob.glob(coupDir + "/data/*tm")
    for file in files:
        name = os.path.basename(file)
        timeVal = float(name.split("_")[-2])
        if timeVal == startTime:
            #startTimeの場合は、rename
            os.rename(file, file[:-2])
        else:
            #以外は、削除
            os.remove(file)

#
#  getFaceCenterLocs
#--------------------
def getFaceCenterLocs(faceCenterDict):
    """ faceCenterDictからfaceCenterLocsを取得して返す"""
    faceNos = list(faceCenterDict.keys())
    faceNos.sort()
    faceCenterLocs = []
    for faceNo in faceNos:
        loc = faceCenterDict[faceNo]
        strLoc = list(map(str, loc))
        faceCenterLocs.append(strLoc)
    return faceCenterLocs
    
#
#  getDictsFromFaceCenterDict
#-----------------------------
def getDictsFromFaceCenterDict(faceCenterDict, allFaceDict, allNodeDict):
    """ mapPointFaceDictとnodeOrderDictを作成。"""
    mapPointFaceDict = {}
    nodeOrderDict = {}
    order = 0
    for faceNo in faceCenterDict.keys():
        nodeNos = allFaceDict[faceNo]["nodes"]
        for nodeNo in nodeNos:
            if not nodeNo in nodeOrderDict.values():
                loc = allNodeDict[nodeNo]
                mapPointFaceDict[nodeNo] = [loc, nodeNos]
                nodeOrderDict[order] = nodeNo
                order += 1
    return mapPointFaceDict, nodeOrderDict

#
#  getMapFaceCenterDict
#-----------------------
def getMapFaceCenterDict(faceCenterDict, allFaceDict):
    """ face中心にmappingする為のmapFaceCenterDictを作成する。
    key=faceNo, value=[[faceCenterLoc], [nodeNos]]"""
    mapDict = {}
    for faceNo in faceCenterDict.keys():
        loc = faceCenterDict[faceNo]
        nodeNos = allFaceDict[faceNo]["nodes"]
        mapDict[faceNo] = [loc] + [nodeNos]
    return mapDict

#
#  saveFaceNodesPara
#--------------------
def saveFaceNodesPara(faceCenterLocs, mapPointFaceDict, nodeOrderDict, pNo):
    """ couplingで使用するmesh内容を保存する"""
    #faceCenterLocを保存
    saveDir = fluidCaseDir + "/coupling_FrontISTR/data/"
    name = patchName + "_faceCenterLocs_" + pNo
    fileName = saveDir + name
    lines = []
    for strLoc in faceCenterLocs:
        line = " ".join(strLoc) + "\n"
        lines.append(line)
    f = open(fileName, "w"); f.writelines(lines); f.close()

    #mapPointFaceDictを保存
    name = patchName + "_mapPointFaceNodes_" + pNo
    fileName = saveDir + name
    lines = []
    nodeOrders = list(nodeOrderDict.keys())
    nodeOrders.sort()
    num = len(mapPointFaceDict.keys())
    for orderNo in nodeOrders:
        nodeNo = nodeOrderDict[orderNo]
        loc = mapPointFaceDict[nodeNo][0]
        strLoc = list(map(str, loc))
        faceNodeNos = mapPointFaceDict[nodeNo][1]
        strNodeNos = list(map(str, faceNodeNos))
        line = " ".join([str(nodeNo)] + strLoc + strNodeNos) + "\n"
        lines.append(line)
    f = open(fileName, "w"); f.writelines(lines); f.close()
    return

#
#  saveMapFaceCenterDictPara
#----------------------------
def saveMapFaceCenterDictPara(mapFaceCenterDict, pNo):
    """ face中心にmappingする為のdataを保存する"""
    saveDir = fluidCaseDir + "/coupling_FrontISTR/data/"
    name = patchName + "_mapFaceCenterNodes_" + pNo
    fileName = saveDir + name
    lines = []
    for faceNo in mapFaceCenterDict.keys():
        loc, nodeNos = mapFaceCenterDict[faceNo]
        words = map(str, [faceNo] + loc + nodeNos)
        line = " ".join(words) + "\n"
        lines.append(line)
    f = open(fileName, "w"); f.writelines(lines); f.close()

#------------------
#  createPatchCont
#------------------
def createPatchCont():
    """ OpenFOAM側のfaceCenterLocs（face中心座標）などmesh情報を取得"""
    if nProcs == "1":
        #singleコアの場合
        fluidCase = pyTreeFoam.case(fluidCaseDir)
        time = fluidCase.getCurrTimeFolder()
        region= "."
        pMesh = fluidCase.getCurrMeshDir(time, region, "boundary")
        meshDir = fluidCaseDir + os.sep + pMesh.split(os.sep)[0]
        polyMesh = createPtc.ofMesh(fluidCaseDir, meshDir)
        faceCenterDict = polyMesh.getFaceCenterDict([patchName])
        #faceCenter座標を取得
        faceCenterLocs = getFaceCenterLocs(faceCenterDict)
        #mapPointFaceDict, nodeOrderDictを作成
        allFaceDict = polyMesh.faceDict()
        allNodeDict = polyMesh.nodeDict()
        mapPointFaceDict, nodeOrderDict = getDictsFromFaceCenterDict(
            faceCenterDict, allFaceDict, allNodeDict)
        #取得したmesh情報を保存
        saveFaceNodesPara(faceCenterLocs, mapPointFaceDict, nodeOrderDict, "0")
        #face中心にmappingする為のdata作成
        mapFaceCenterDict = getMapFaceCenterDict(faceCenterDict, allFaceDict)
        #ファイルに保存
        saveMapFaceCenterDictPara(mapFaceCenterDict, "0")
    else:
        #parallelの場合
        fluidParaCase = pyTreeFoam.paraCase(fluidCaseDir)
        time = fluidParaCase.getCurrTimeFolder()
        region= "."
        pMesh = fluidParaCase.getCurrMeshDir(time, region, "boundary")
        
        for p in range(int(nProcs)):
            pNo = str(p)
            meshDir = fluidCaseDir + "/processor" + pNo + "/" + pMesh.split(os.sep)[0]
            polyMesh = createPtc.ofMesh(fluidCaseDir + "/processor" + pNo, meshDir)
            faceCenterDict = polyMesh.getFaceCenterDict([patchName])
            #faceCenter座標を取得
            faceCenterLocs = getFaceCenterLocs(faceCenterDict)
            #mapPointFaceDict, nodeOrderDictを作成
            allFaceDict = polyMesh.faceDict()
            allNodeDict = polyMesh.nodeDict()
            mapPointFaceDict, nodeOrderDict = getDictsFromFaceCenterDict(
                faceCenterDict, allFaceDict, allNodeDict)
            saveFaceNodesPara(faceCenterLocs, mapPointFaceDict, nodeOrderDict, pNo)
            mapFaceCenterDict = getMapFaceCenterDict(faceCenterDict, allFaceDict)
            saveMapFaceCenterDictPara(mapFaceCenterDict, pNo)

#
#  getNeedNodesFromDict
#-----------------------
def getNeedNodesFromDict(sgrpDict, nodeDict):
    """ 必要なnodeを抽出したnodeDictを取得して返す。"""
    newNodeDict = {}
    for key in sgrpDict.keys():
        nodeNos = sgrpDict[key][1]
        for nodeNo in nodeNos:
            if not nodeNo in newNodeDict.keys():
                newNodeDict[nodeNo] = nodeDict[nodeNo]
    return newNodeDict

#
#  saveFistrFaceData
#--------------------
def saveFistrFaceData(sgrpDict, nodeDict):
    """ FrontISTR側のfaceDataを保存する"""
    #sgrpDictを保存
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    name = sgrpName + "_mapFaceCenterFaceNodes"
    fileName = dataDir + "/" + name
    lines = []
    for key in sgrpDict.keys():
        elmNo, faceNo = key
        loc = sgrpDict[key][0]
        nodeNos = sgrpDict[key][1]
        nums = [elmNo, faceNo] + loc + nodeNos
        line = " ".join(map(str, nums)) + "\n"
        lines.append(line)
    f = open(fileName, "w"); f.writelines(lines); f.close()
    #nodeDictを保存
    name = sgrpName + "_nodeLocs"
    fileName = dataDir + "/" + name
    lines = []
    for nodeNo in nodeDict.keys():
        loc = nodeDict[nodeNo]
        nums = [nodeNo] + loc
        line = " ".join(map(str, nums)) + "\n"
        lines.append(line)
    f = open(fileName, "w"); f.writelines(lines); f.close()

#-----------------------
#  createSgrpCont
#-----------------------
def createSgrpCont():
    """ FrontISTR側のSGRPの情報を取得して保存する"""
    mesh = mapping.fistrMesh(solidCaseDir)
    meshHeaderData = mesh.getMeshHeaderData()
    nodeDict = mesh.nodeDict(meshHeaderData)
    elementDict = mesh.elementDict(meshHeaderData)
    sgrpData = mesh.getSgrpData(sgrpName, meshHeaderData)
    sgrpDict = mesh.getSgrpDict(sgrpData, nodeDict, elementDict)
    newNodeDict = getNeedNodesFromDict(sgrpDict, nodeDict)
    #Dictを保存
    saveFistrFaceData(sgrpDict, newNodeDict)
    return (sgrpDict, newNodeDict)

#
#  clearRestartSetting
#----------------------
def clearRestartSetting():
    """ FistrMode.cntファイルのresartの設定をクリアする"""
    if fsi == "yes":
        fileName = solidCaseDir + "/FistrModel_fsi.cnt"
        f = open(fileName); lines = f.readlines(); f.close()
        #restartの設定
        i = pyFistr.getNextHeaderName(lines, 0, "!RESTART", "")
        lines[i] = pyFistr.setValName(lines[i], "FREQUENCY=1")
        #nStepsをクリア
        i = pyFistr.getNextHeaderName(lines, 0, "!DYNAMIC", "")
        i += 2
        if i >= len(lines):
            print("error: it is not DYNAMIC analysis. must set at DYNMIC.")
            print()
            return
        words = pyFistr.deleteSp(lines[i]).split(",")
        words[2] = nStepsOF
        lines[i] = ", ".join(words) + "\n"
        #保存
        f = open(fileName, "w"); f.writelines(lines); f.close()
    if cht == "yes":
        fileName = solidCaseDir + "/FistrModel_cht.cnt"
        f = open(fileName); lines = f.readlines(); f.close()
        #restartの設定
        i = pyFistr.getNextHeaderName(lines, 0, "!RESTART", "")
        lines[i] = pyFistr.setValName(lines[i], "FREQUENCY=1")
        #nStepsをクリア
        i = pyFistr.getNextHeaderName(lines, 0, "!HEAT", "")
        i += 1
        if i >= len(lines):
            print("error: it is not HEAT analysis. must set at HEAT.")
            print()
            return
        words = pyFistr.deleteSp(lines[i]).split(",")
        num = int(nStepsOF) * float(deltaT)
        #endTime = roundVal(num, 8)
        endTime = num
        words[1] = str(endTime)
        lines[i] = ", ".join(words) + "\n"
        #保存
        f = open(fileName, "w"); f.writelines(lines); f.close()
    #結果fileを全て削除
    delFiles = glob.glob(solidCaseDir + "/*.pvtu")
    delFiles += glob.glob(solidCaseDir + "/FistrModel.res.*")
    delFiles += glob.glob(solidCaseDir + "/FistrModel_temp.res.*")
    delFiles += glob.glob(solidCaseDir + "/*.particles")
    delFiles += glob.glob(solidCaseDir + "/*.vtk")
    delFiles += glob.glob(solidCaseDir + patchName + "_faceCenterPress*.particles")
    for delFile in delFiles:
        os.remove(delFile)
    delFolders = glob.glob(solidCaseDir + "/FistrModel.vis_psf.*")
    for delFolder in delFolders:
        if os.path.isdir(delFolder) == True:
            shutil.rmtree(delFolder)

#
#  createCntFile
#----------------
def createCntFile():
    """ 計算に必要なdat, cntファイルを作成する。"""
    datFsiFile = solidCaseDir + os.sep + "hecmw_ctrl_fsi.dat"
    datChtFile = solidCaseDir + os.sep + "hecmw_ctrl_cht.dat"
    datChtssFile = solidCaseDir + os.sep + "hecmw_ctrl_chtss.dat"
    cntMshFiles = pyCoupling.cntMshFiles(solidCaseDir)
    #fsiの連成？
    if fsi == "yes":
        cntMshFiles.checkFsiFiles()
        remakeMaterialValues(datFsiFile)
    #chtの連成？
    if cht == "yes":
        cntMshFiles.checkChtFiles()
        createInitialTemperatureFile()
        remakeMaterialValues(datChtFile)
    #chtssの連成
    if chtss == "yes":
        cntMshFiles.checkChtssFiles()
        remakeTemperatureCard()
        remakeMaterialValues(datChtssFile)

#
#  createInitialTemperatureFile
def createInitialTemperatureFile():
    """ solidの初期温度を保存する。"""
    #mshFile = solidCaseDir + os.sep + "FistrModel.msh"
    cntName, mshName = pyFistr.cntMshFileName(solidCaseDir)
    mshFile = solidCaseDir + os.sep + mshName
    f = open(mshFile); lines = f.readlines(); f.close()
    headerData = pyFistr.getHeaderData(lines)
    tempLines = []
    for i in range(len(headerData)):
        header = headerData[i][0]
        if header[:len("!INITIAL CONDITION")] == "!INITIAL CONDITION":
            tempLines = headerData[i][1]
            break
    if len(tempLines) == 0:
        tempLines = ["ALL, 300.0\n"]
    elif len(tempLines) > 1:
        #複数行のinitialTempの場合、直ぐに戻る
        return
    iniTempFile = solidCaseDir + os.sep + "initialTemperature"
    f = open(iniTempFile, "w"); f.writelines(tempLines); f.close()

#
#  remakeTemperatureCard
def remakeTemperatureCard():
    """ !TEMPERATURE ALLを整形する。"""
    cntFile = solidCaseDir + os.sep + "FistrModel_chtss.cnt"
    f = open(cntFile); lines = f.readlines(); f.close()
    i = 0
    si = 0
    while i < len(lines):
        i = pyFistr.getNextHeaderName(lines, i, "!TEMPERATURE", "")
        i += 1
        if i < len(lines):
            words = pyFistr.deleteSp(lines[i]).split(",")
            if words[0] == "ALL":
                si = i + 1
                ei = pyFistr.skipNextHeader(lines, i)
                break
    #!TEMPERATURE ALLがあるか？
    if si > 0:
        #!TEMPERATURE cardの'otherData=yes'があれば削除
        line = lines[si-2]
        words = pyFistr.deleteSp(line).split(",")
        newWords = []
        for word in words:
            if word != "otherData=yes":
                newWords.append(word)
        line = ", ".join(newWords) + "\n"
        lines[si-2] = line
        #linesを再作成し、保存
        lines = lines[:si] + lines[ei:]
        f = open(cntFile, "w"); f.writelines(lines); f.close()

#
#  remakeMaterialValues
def remakeMaterialValues(datFile):
    """ chtCntFileの物性値をmshFileにコピーする。"""
    cntName, mshName = getCntMshNameFromDatFile(datFile)
    cntFile = solidCaseDir + os.sep + cntName
    mshFile = solidCaseDir + os.sep + mshName
    remakeMaterialValuesFromCntMsh(cntFile, mshFile)

#
#  getCntMshNameFromDatFile
def getCntMshNameFromDatFile(datFile):
    """ datFileからcnt,msh名を取得して返す"""
    f = open(datFile); lines = f.readlines(); f.close()
    #cntFile
    i = pyFistr.getNextHeaderName(lines, 0, "!CONTROL", "NAME=fstrCNT")
    i += 1
    cntName = pyFistr.deleteSp(lines[i])
    #mshFile
    i = pyFistr.getNextHeaderName(lines, 0, "!MESH", "NAME=fstrMSH")
    i += 1
    mshName = pyFistr.deleteSp(lines[i])
    return cntName, mshName

#
#  remakeMaterialValuesFromCntMsh
def remakeMaterialValuesFromCntMsh(cntName, mshName):
    valLineDict = {}
    f = open(cntName); lines = f.readlines(); f.close()
    #材料名と値を取得
    i = 0
    while i < len(lines):
        i = pyFistr.getNextHeaderName(lines, i, "!MATERIAL", "")
        if i + 6 >= len(lines):
            break
        name = pyFistr.getValName(lines[i], "NAME")
        values = []
        i += 2
        values.append(lines[i])
        i += 2
        values.append(lines[i])
        i += 2
        values.append(lines[i])
        valLineDict[name] = values
    #msh側に設定
    f = open(mshName); lines = f.readlines(); f.close()
    i = 0
    while i < len(lines):
        i = pyFistr.getNextHeaderName(lines, i, "!MATERIAL", "")
        if i + 6 >= len(lines):
            break
        name = pyFistr.getValName(lines[i], "NAME")
        vals = valLineDict[name]
        i += 2
        n = str(len(vals[0].split(",")))
        lines[i-1] = pyFistr.setValName(lines[i-1], "SUBITEM="+str(n))
        lines[i] = vals[0]
        i += 2
        lines[i] = vals[1]
        i += 2
        lines[i] = vals[2]
    f = open(mshName, "w"); f.writelines(lines); f.close()

#
#  setInitialSettingFistr
#-------------------------
def setInitialSettingFistr():
    """ Fistr側の初期設定"""
    #step時間を取得
    dT = float(deltaT)
    #step数とstep時間を設定（時間増分）
    if fsi == "yes":
        couplingInterval = dT * int(nStepsOF)
        stepTime = str(couplingInterval / int(nStepsOF))
        cntFile = solidCaseDir + os.sep + "FistrModel_fsi.cnt"
        #cntFileを修正
        fileName = cntFile
        f = open(fileName); lines = f.readlines(); f.close()
        i = pyFistr.getNextHeaderName(lines, 0, "!DYNAMIC", "")
        i += 2
        if i >= len(lines):
            i = pyFistr.getNextHeaderName(lines, 0, "!SOLUTION", "")
            lines[i] = pyFistr.setValName(lines[i], "TYPE=DYNAMIC")
            i = pyFistr.getNextHeaderName(lines, 0, "!HEAT", "")
            lines[i] = "!DYNAMIC, TYPE=NONLINEAR\n"
            si = i
            header = lines[:i+1]
            i = pyFistr.skipNextHeader(lines, i+1)
            footer = lines[i:]
            #時間増分:0.0001, 全step数:2 を設定
            contLines  = ["1, 1\n"]
            contLines += ["0.0, 1.0, 2, 0.0001\n"]
            contLines += ["0.5, 0.25\n"]
            contLines += ["1, 1, 0.01, 0.01\n"]
            contLines += ["100, 1, 1\n"]
            contLines += ["0, 0, 0, 0, 0, 0\n"]
            lines = header + contLines + footer
            i = si + 2
        words = pyFistr.deleteSp(lines[i]).split(",")
        words[2] = nStepsOF       #全step数
        words[3] = stepTime     #時間増分
        lines[i] = ", ".join(words) + "\n"
        #結果の書き込み頻度を設定
        fr = str(int(nStepsOF) * nSubSteps)
        i = pyFistr.getNextHeaderName(lines, 0, "!WRITE", "")
        lines[i] = pyFistr.setValName(lines[i], "FREQUENCY="+fr)
        i = pyFistr.getNextHeaderName(lines, i+1, "!WRITE", "")
        lines[i] = pyFistr.setValName(lines[i], "FREQUENCY="+fr)
        #cntファイル書き込み
        fw = open(fileName, "w"); lines = fw.writelines(lines); fw.close()
    if cht == "yes":
        couplingInterval = dT * int(nSteps)
        stepTime = str(couplingInterval / int(nSteps))
        cntFile = solidCaseDir + os.sep + "FistrModel_cht.cnt"
        mshFile = solidCaseDir + os.sep + "FistrModel_cht.msh"
        #cntFileを修正
        fileName = cntFile
        f = open(fileName); lines = f.readlines(); f.close()
        i = pyFistr.getNextHeaderName(lines, 0, "!HEAT", "")
        i += 1
        if i >= len(lines):
            i = pyFistr.getNextHeaderName(lines, 0, "!SOLUTION", "")
            lines[i] = pyFistr.setValName(lines[i], "TYPE=HEAT")
            i = pyFistr.getNextHeaderName(lines, 0, "!DYNAMIC", "")
            si = i
            header = lines[:i]
            i = pyFistr.skipNextHeader(lines, i+1)
            footer = lines[i:]
            #時間増分:0.0001, 終了時間:0.0002(全step数:2) を設定
            solLines  = "!HEAT\n"
            solLines += "0.0001, 0.0002, 0, 20.0, 20, 1.0e-6\n"
            lines = header + solLines + footer
            i = si + 1
        words = pyFistr.deleteSp(lines[i]).split(",")
        endTime = str(couplingInterval)
        words[0] = stepTime     #時間増分
        words[1] = endTime      #終了時間
        lines[i] = ", ".join(words) + "\n"
        #結果の書き込み頻度を設定
        fr = nSteps
        #fr = nStepsOF
        i = pyFistr.getNextHeaderName(lines, 0, "!WRITE", "")
        lines[i] = pyFistr.setValName(lines[i], "FREQUENCY="+fr)
        i = pyFistr.getNextHeaderName(lines, i+1, "!WRITE", "")
        lines[i] = pyFistr.setValName(lines[i], "FREQUENCY="+fr)
        #cntファイル書き込み
        fw = open(fileName, "w"); lines = fw.writelines(lines); fw.close()
        #mshファイルのINITIAL_CONDITIONの初期設定
        iniTempFile = solidCaseDir + os.sep + "initialTemperature"
        f = open(iniTempFile); tempLines = f.readlines(); f.close()
        fileName = mshFile
        f = open(fileName); lines = f.readlines(); f.close()
        headerData = pyFistr.getHeaderData(lines)
        for i in range(len(headerData)):
            header = headerData[i][0]
            if header[:len("!INITIAL CONDITION")] == "!INITIAL CONDITION":
                headerData[i][1] = tempLines
                break
        #mshファイル書込
        lines = []
        for header in headerData:
            lines += [header[0]]
            lines += header[1]
        f = open(fileName, "w"); f.writelines(lines); f.close()

#----------------
#  runCoupling
#----------------
def runCoupling(sgrpDict, nodeDict):
    """ FrontISTRの起動"""
    global beforeTime, currTime
    global mappingTriDict, deltaT, procOF
    global fsiDict, chtDict
    #変数の初期化
    fsiDict = {}
    chtDict = {}
    #openFOAMの起動
    procOF = runOpenFOAM()
    mappingTriDict = {}
    #Fistrを起動する準備
    skipSP = int(nStepsOF)      #fsiのnStep数
    skipST = int(nSteps)        #temp関連のnStep数
    skipSTS = 3                 #chtssのnStep数
    fsiDict["beforeCouplingT"] = float(deltaT) * int(nStepsOF)
    chtDict["beforeCouplingT"] = float(deltaT) * int(nSteps)
    fsiDict["beforeCurrTime"] = currTime
    chtDict["beforeCurrTime"]  = currTime
    fsiDict["nSteps"] = nStepsOF
    chtDict["nSteps"] = nSteps
    fsiDict["startStep"] = runStepFsi
    chtDict["startStep"] = runStepCht
    fsiDict["runStep"] = runStepFsi
    chtDict["runStep"] = runStepCht
    fsiDict["beforeStartStep"] = runStepFsi
    chtDict["beforeStartStep"] = runStepCht
    fsiDict["corrStep"] = 0
    chtDict["corrStep"] = 0
    #開始時間のstr文字を取得(restart時間含む)
    startTimeStr = getStartTimeOfString()
    delTimes.append(startTimeStr)
    #初期の温度dataを作成
    if couplingDict["cht"] == "yes":
        #FrontIstrが使う初期温度
        tempLines = createInitialTemp(startTimeStr)
        #OpenFOAMが読み込むfile
        createInitialTempFiles(nProcs, startTimeStr, tempLines)
    #初期の変位、圧力dataFileを作成
    if couplingDict["fsi"] == "yes" or couplingDict["chtss"] == "yes":
        if float(couplingDict["startTime"]) == 0.0:
            #初期から開始の場合
            createInitialDispFiles(nProcs)
        else:
            #途中から再開の場合
            createRestartDispFiles(nProcs, startTimeStr)        

    #変数の初期化
    if fsi == "yes":
        runStep = runStepFsi
    else:
        runStep = runStepCht
    delSteps.append(str(runStep))
    iniFlag = 0
    loop = True
    chtssDispPtc = []
    fsiDispPtc = []
    sendDispSteps = []
    while loop: 
        #次のcoulingまで待つ
        (beforeTime, currTime, statOF,
         skipSP, skipST, skipSTS, runStep) = getNextCouplingTime(procOF, runStep)     #OF停止もチェック
        #openFOAMの生死確認
        if statOF == "halt":
            break

        #リスタートファイルを作成
        createRestartFileForFistr(currTime, runStep)
        #endTime以内？
        if (float(currTime) < float(couplingDict["endTime"])):
        
            #fsi？
            if fsi == "yes" and skipSP == 0:
                #fsi用のdatFileをセット
                setFsiChtCntFile("fsi")
                #時間増分を更新
                deltaT = str(float(currTime) - float(beforeTime))
                fsiDict = setDeltaTInCnt_fsi(fsiDict, currTime, deltaT, iniFlag, runStep)                 
                startStep = fsiDict["startStep"]
                endStep = fsiDict["endStep"]
                endTime = float(currTime) + (endStep - runStep) * float(deltaT)
                #圧力をFistr側にmappingする
                mappingPressPtcFile(sgrpDict, nodeDict, iniFlag)
                print("(py) FrontISTR is running for structure from step " + str(startStep) +
                        " to " + str(endStep) + " time " + str(endTime) + 
                        " ...", flush=True)
                #Fistrを起動
                runFistr("fsi", endStep)
                print("(py) displacement PTC of pressure is creating...", flush=True)
                #変位のptcデータを作成
                fsiDispPtc = createFsiDispPtc(nodeDict)
                #変位のendStepを保存
                if not endStep in sendDispSteps:
                    sendDispSteps.append(endStep)
                    sendDispSteps.sort()
            #cht？
            if cht == "yes" and skipST == 0:
                #cht用のdatFileをセット
                setFsiChtCntFile("cht")
                #時間増分を更新
                deltaT = str(float(currTime) - float(beforeTime))
                chtDict = setDeltaTInCnt_cht(chtDict, currTime, deltaT, iniFlag, runStep)
                startStep = chtDict["startStep"]
                endStep = chtDict["endStep"]
                endTime = float(currTime) + (endStep - runStep) * float(deltaT)
                #heatFluxを取得
                mappingHtPtcFile(sgrpDict, nodeDict, iniFlag)
                print("(py) FrontISTR is running for heat transfer from step " + str(startStep) +
                        " to " + str(endStep) + " time " + str(endTime) + 
                        "...", flush=True)
                #tempLinesをmshFileに設定
                #Fistrを起動
                runFistr("cht", endStep)
                #mshFileを更新
                tempLines = addTempRenameRes(runStep, endStep, currTime)
                #chtss?
                if chtss == "yes" and skipST == 0:
                    #chtss用のdatFileをセット
                    setFsiChtCntFile("chtss")
                    #solidに温度を設定
                    remakeFistrCntFileForTemperature(tempLines)
                    print("(py) FrontISTR is running for heat stress...", flush=True)
                    #Fistrを起動
                    runFistr("chtss", endStep)
                    print("(py) displacement PTC of heat stress is creating...", flush=True)
                    #熱変位のptcデータを作成
                    chtssDispPtc = createChtssDispPtc(nodeDict)
                    #変位のendStepを保存
                    if not endStep in sendDispSteps:
                        sendDispSteps.append(endStep)
                        sendDispSteps.sort()
            #disp転送？
            if ((chtss == "yes" and skipST == 0) or 
                (fsi == "yes" and skipSP == 0)):
                #変位ptcデータを作成
                dispPtc = addDisplacements(chtssDispPtc, fsiDispPtc)
                #dispPtcを転送
                print("(py) displacement value is sending to OpenFOAM...")
                dispConts = createDispPtcFileFromPtc(dispPtc)
                sendStep = sendDispSteps.pop(0)
                title = "displacement " + str(sendStep)
                sendContsToStream(title, dispConts)

            #temp転送？
            if cht == "yes" and skipST == 0:
                #温度のptcを作成し転送
                print("(py) temperature PTC of heat flux is creating...", flush=True)
                temp = createTempPtcFile(nodeDict, tempLines)
                print("(py) temperature value is sending to OpenFOAM...")
                title = "temperature"
                sendContsToStream(title, temp)
            iniFlag = 1
            #結果ファイルを削除
            clearPressPtcResFile(beforeTime, currTime, skipSP, skipST, runStep)

    renameTemporaryFile()
    mmTimes.flush()
    mmTimes.close()
    mmCoupling.flush()
    mmCoupling.close()

#
#  setFsiChtCntFile
def setFsiChtCntFile(sect):
    """ sectに応じたdatファイルに修正"""
    pasteDatFile = solidCaseDir + os.sep + "hecmw_ctrl.dat"
    if sect == "fsi":
        datFile = solidCaseDir + os.sep + "hecmw_ctrl_fsi.dat"
    elif sect == "cht":
        datFile = solidCaseDir + os.sep + "hecmw_ctrl_cht.dat"
    elif sect == "chtss":
        datFile = solidCaseDir + os.sep + "hecmw_ctrl_chtss.dat"
    shutil.copy(datFile, pasteDatFile)

#
#  getStartTimeOfString
#------------------------
def getStartTimeOfString():
    """ startTimeを8桁のstr文字を取得"""
    startTimeStr = "0.000000"
    num = float(couplingDict["startTime"])
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    if couplingDict["fsi"] == "yes":
        pattern = dataDir + "/" + patchName + "_facePress_*"
    else:
        pattern = dataDir + "/" + patchName + "_faceHeatFlux_*"
    dispFiles = glob.glob(pattern)
    for dispFile in dispFiles:
        name = os.path.basename(dispFile)
        word = name.split("_")[-2]
        wordNum = float(word)
        if num == wordNum:
            startTimeStr = word
            break
    return startTimeStr

#
#  createInitialTemp
#---------------------
def createInitialTemp(startTimeStr):
    """ FrontISTRが使う計算初期のtempLinesを取得する"""
    startStep = chtDict["startStep"]
    #温度dataを作成
    if float(startStep) == 0.0:
        #初回
        fileName = solidCaseDir + os.sep + "initialTemperature"
        f = open(fileName); tempLines = f.readlines(); f.close()
    else:
        #途中から再開
        fileName = solidCaseDir + os.sep + "FistrModel_temp.res.0." + str(startStep)
        tempLines = getTempLinesInResFile(fileName)
    return tempLines

#
#  createInitialTempFiles
#-------------------------
def createInitialTempFiles(nProcs, startTimeStr, tempLines):
    """ OpenFOAMが読み込む初期温度fileを作成する"""
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    iniTemp = tempLines[0].split(",")[1]
    if float(couplingDict["startTime"]) == 0.0:
        #計算初回の温度を設定
        for i in range(int(nProcs)):
            procNo = str(i)
            name = patchName + "_mapPointFaceNodes_" + procNo
            fileName = dataDir + "/" + name
            f = open(fileName); lines = f.readlines(); f.close()
            #保存のfileName
            name = patchName + "_temp_0_" + procNo + "tm"
            tempFile = dataDir + "/" + name
            dataLine = iniTemp + "\n"
            f = open(tempFile, "w")
            for i in range(len(lines)):
                f.write(dataLine)
            f.close()
    else:
        #計算再開時の温度を設定
        pattern = dataDir + "/" + patchName + "_temp_" + startTimeStr + "_*"
        fileNames = getAllProcsFiles(pattern, int(nProcs))
        for fileName in fileNames:
            if os.path.exists(fileName):
                os.rename(fileName, fileName + "tm")

#
#  createInitialDispFiles
#-------------------------
def createInitialDispFiles(nProcs):
    """ 初期の変位dataを作成し、セットする"""
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    for i in range(int(nProcs)):
        procNo = str(i)
        name = patchName + "_mapPointFaceNodes_" + procNo
        fileName = dataDir + "/" + name
        f = open(fileName); lines = f.readlines(); f.close()
        #保存のfileName
        endStep = "0"
        name = patchName + "_disp_" + endStep + "_" + procNo + "tm"
        dispFile = dataDir + "/" + name
        dataLine = "0.0 0.0 0.0\n"
        f = open(dispFile, "w")
        for ii in range(len(lines)):
            f.write(dataLine)
        f.close()
        #圧力の初期値作成
        procNo = str(i)
        name = patchName + "_facePress_0.000000_" + procNo + "tm"
        pressFile = dataDir + "/" + name
        dataLine = "0.0\n"
        f = open(pressFile, "w")
        for ii in range(len(lines)):
            f.write(dataLine)
        f.close()

#
#  createRestartDispFiles
#-------------------------
def createRestartDispFiles(nProcs, startTimeStr):
    """ 再スタート時の圧力dataを作成する"""
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    pattern = dataDir + "/" + patchName + "_facePress_" + startTimeStr + "_*"
    fileNames = getAllProcsFiles(pattern, int(nProcs))
    for fileName in fileNames:
        if os.path.exists(fileName):
            os.rename(fileName, fileName + "tm")

#
#  addTempRenameRes
#-------------------
def addTempRenameRes(runStep, endStep, currTime):
    """ mshFileに温度の計算結果を追加してrenameする"""
    #温度データを取得
    nSteps = int(chtDict["endTime"] / float(couplingDict["deltaT"]) + 0.5)
    resFile = solidCaseDir + os.sep + "FistrModel_temp.res.0." + str(nSteps)
    tempLines = getTempLinesInResFile(resFile)
    deltaT = str(float(currTime) - float(beforeTime))
    #nextTime = endStep * float(deltaT)
    nextTime = float(currTime) + (endStep - chtDict["startStep"]) * float(deltaT)
    if pyCoupling.isWriteInterval(nextTime, writeInterval, deltaT) == True:
        #温度データをrename
        #newName = solidCaseDir + os.sep + "FistrModel_temp.res.0." + str(currTime)
        newName = solidCaseDir + os.sep + "FistrModel_temp.res.0." + str(endStep)
        if resFile != newName:
            os.rename(resFile, newName)
    #meshFileに温度データを初期値として追加
    addTempDataToMeshFile(tempLines)
    return tempLines

#
#  getTempLinesInResFile
def getTempLinesInResFile(resFile):
    """ resFileから温度データを取得する"""
    tempLines = []
    f = open(resFile); lines = f.readlines(); f.close()
    flag = 0
    i = 0
    while i < len(lines):
        if flag == 1:
            nodeNo = lines[i].split()[0]
            temp = lines[i+1].split()[0]
            tempLines.append(", ".join([nodeNo, temp]) + "\n")
            i += 1
        elif lines[i][:len("TEMPERATURE")] == "TEMPERATURE":
            flag = 1
        i += 1
    return tempLines

#
#  addTempDataToMeshFile
def addTempDataToMeshFile(tempLines):
    """ mshFileに温度データを追加"""
    mshFile = solidCaseDir + os.sep + "FistrModel_cht.msh"
    f = open(mshFile); lines = f.readlines(); f.close()
    headerData = pyFistr.getHeaderData(lines)
    for i in range(len(headerData)):
        header = headerData[i][0]
        if header[:len("!INITIAL CONDITION")] == "!INITIAL CONDITION":
            headerData[i][1] = tempLines
            break
    lines = []
    for header in headerData:
        lines += [header[0]]
        lines += header[1]
    f = open(mshFile, "w"); f.writelines(lines); f.close()

#
#  addTimeToTimeList
def addTimeToTimeList(bTime):
    """ timeListにtimeを追加する"""
    mmTimeList.seek(0)
    conts = mmTimeList.read().decode()
    lines = conts.split("\n")
    lines = [bTime] + lines
    lines = lines[:5]
    conts = "\n".join(lines) + "\n"
    conts = conts[:200]
    mmTimeList.seek(0)
    mmTimeList.write(conts.encode())
    mmTimeList.flush()        

#
#  getNextCouplingTime
#-----------------------
def getNextCouplingTime(procOF, runStep):
    """ 次のbeforeTime, currTimeを取得する"""
    global delTimes, delSteps
    statOF = "run"
    #counterをリセットして、更新内容を取得
    count = waitCount
    sleepTime = waitTime
    xTime = beforeTime
    bTime = beforeTime; cTime = currTime
    couplePress = "9"
    coupleTemp = "9"
    coupleOther = "9"
    while True:
        #時間を読み込み
        timesDict = pyCoupling.getTimeCont(mmTimes)
        if len(timesDict) >= 4:
            bTime = timesDict["beforeTime"]
            cTime = timesDict["currTime"]
            #timeが更新？
            if float(xTime) != float(bTime):
                xTime = bTime
                delTimes.append(cTime)
                #runStepを更新
                runStep = int(timesDict["timeIndex"])
                delSteps.append(str(runStep))
                #timeList(共有メモリ）にbeforeTimeを追加
                addTimeToTimeList(bTime)

                #skipStepsを設定
                couplePress = int(timesDict["couplePress"])
                coupleTemp = int(timesDict["coupleTemp"])
                coupleOther = int(timesDict["coupleOther"])
                #mmCouplingに保存
                pyCoupling.writeTimeCont(mmCoupling, timesDict)
                break
                
        #counter以内？
        count -= 1
        if count < 0:
            break
        #OFの生死を確認
        if procOF.poll() != None:
            statOF = "halt"
            break
        #待ち
        time.sleep(sleepTime)
    return bTime, cTime, statOF, couplePress, coupleTemp, coupleOther, runStep

#
#  getStreamEachProcs
#--------------------
def getStreamEachProcs(procNo, q):
    """ 並列処理用のプロセス。並列数分起動される。
    自身のstreamから内容を取り出す。"""
    lines = getProcStreamLines(procNo, "pressurePtcData")
    q.put(lines)

#
#  mappingPressPtcFile
#-----------------
def mappingPressPtcFile(sgrpDict, nodeDict, iniFlag):
    """ 散らばっている圧力ptcをまとめ、Fistr側にmappingする。"""
    global mappingTriDict
    #stream内にある圧力ptcデータを取得
    #  並列数分の読み込みprocessを起動する
    print("(py) pressure value is receiving from OpenFOAM...", flush=True)
    queue = []
    procs = []
    for i in range(int(nProcs)):
        q = multiprocessing.Queue()
        queue.append(q)
        p = multiprocessing.Process(
            target=getStreamEachProcs, args=(str(i), q))
        p.start()
        procs.append(p)
    #  全steamの内容を取得
    lines = []
    for i in range(int(nProcs)):
        lines += queue[i].get()
    #fileに保存
    name = patchName + "_faceCenterPress.particles"
    allPtcFile = solidCaseDir + "/" + name
    f = open(allPtcFile, "w"); f.writelines(lines); f.close()
    #ptcData(数値data）を取得
    mesh = mapping.fistrMesh(solidCaseDir)
    ptcData = mesh.getPtcDataFromLines(lines)
    if len(ptcData) < 3:
        return "error could not get pressure from point cloud."
    #初回計算？
    if iniFlag == 0:
        #mapping用の3点を取得
        mapStat, mappingTriDict = mesh.getMappingTriLocs(ptcData, sgrpDict, nodeDict)
        if mapStat != "OK":
            print("error: in paressure mapping. small edge may be more finer.")
            return
    #圧力をmapping
    print("(py) pressure value is mapping to SGRP '" + sgrpName + "' of FrontISTR...", flush=True)
    mappingDict = mesh.setMappingDataFromTriLocs(ptcData, sgrpDict, mappingTriDict)
    #mapping結果を保存
    lines = mesh.createMappingFile(mappingDict, sgrpName, allPtcFile)
    name = "mappingSP_" + sgrpName + ".map"
    fileName = solidCaseDir + os.sep + name
    f = open(fileName, "w"); f.writelines(lines); f.close()
    #mapping結果をcntファイルに反映
    remakeFistrCntFileForPress(lines, iniFlag)

#
#  mappingHtPtcFile
#-------------------
def mappingHtPtcFile(sgrpDict, nodeDict, iniFlag):
    """ 散らばっているheatFluxPtcをまとめ、Fistr側にmappingする。"""
    global mappingTriDict
    #stream内にあるhtPtcデータを取得
    #  並列数分の読み込みprocessを起動する
    print("(py) heat flux value is receiving from OpenFOAM...", flush=True)
    queue = []
    procs = []
    for i in range(int(nProcs)):
        q = multiprocessing.Queue()
        queue.append(q)
        p = multiprocessing.Process(
            target=getStreamEachProcs, args=(str(i), q))
        p.start()
        procs.append(p)
    #  全steamの内容を取得
    lines = []
    for i in range(int(nProcs)):
        lines += queue[i].get()
    #fileに保存
    name = patchName + "_faceCenterHeatFlux.particles"
    allPtcFile = solidCaseDir + "/" + name
    f = open(allPtcFile, "w"); f.writelines(lines); f.close()
    #ptcData(数値data）を取得
    mesh = mapping.fistrMesh(solidCaseDir)
    ptcData = mesh.getPtcDataFromLines(lines)
    if len(ptcData) < 3:
        return "error could not get heat flux from point cloud."
    mapStat, mappingTriDict = mesh.getMappingTriLocs(ptcData, sgrpDict, nodeDict)
    if mapStat != "OK":
        print("error: in heatFlux mapping. small edge may be more finer.")
        return
    #圧力をmapping
    print("(py) heat flux value is mapping to SGRP '" + sgrpName + "' of FrontISTR...", flush=True)
    mappingDict = mesh.setMappingDataFromTriLocs(ptcData, sgrpDict, mappingTriDict)
    #mapping結果を保存
    lines = mesh.createMappingFile(mappingDict, sgrpName, allPtcFile)
    name = "mappingSHF_" + sgrpName + ".map"
    fileName = solidCaseDir + os.sep + name
    f = open(fileName, "w"); f.writelines(lines); f.close()
    #mapping結果をcntファイルに反映
    remakeFistrCntFileForHeatFlux(lines)

#
#  remakeFistrCntFileForHeatFlux
#---------------------------------
def remakeFistrCntFileForHeatFlux(mapLines):
    """ heatFluxのmapping結果をcntファイルに反映"""
    addLines = []
    for line in mapLines:
        if line[0] != "#":
            words = line.split()
            if len(words) >= 3:
                words[1] = "S" + words[1]
                newLine = ", ".join(words) + "\n"
                addLines.append(newLine)
    #cntFileを修正
    cntFile = solidCaseDir + os.sep + "FistrModel_cht.cnt"
    f = open(cntFile); lines = f.readlines(); f.close()
    #挿入場所は、最初の!DFLUXの該当SGRPに挿入する。
    i = 0
    while i < len(lines):
        i = pyFistr.getNextHeaderName(lines, i, "!DFLUX", "")
        i += 1
        if i >= len(lines):
            mess = "error: '" + sgrpName + "' does not exist in !DFLUX.\n"
            print(mess)
            si = i
            break
        words = pyFistr.deleteSp(lines[i]).split(",")
        if words[0] == sgrpName:
            si = i - 1
            ei = pyFistr.skipNextHeader(lines, i)
            break
    if si < len(lines):
        #cntFile修正、書き込み
        name = "mappingSP_" + sgrpName + ".map"
        grpid = pyFistr.getValName(lines[si], "GRPID")
        words = ["!DFLUX", "GRPID="+grpid, "forceType=mapping", "mapFile="+name,
                 "mapCoeff=1.0"]
        lines[si] = ", ".join(words) + "\n"
        lines[si+1] = sgrpName + ", S1, 0.0\n"
        newLines = lines[:si+2] + addLines + lines[ei:]
        f = open(cntFile, "w"); f.writelines(newLines); f.close()

#
#  remakeFistrCntFileForTemperature
#-----------------------------------
def remakeFistrCntFileForTemperature(tempLines):
    """ temperatureの計算結果をcntファイルに反映"""
    #cntFileを修正
    cntFile = solidCaseDir + os.sep + "FistrModel_chtss.cnt"
    f = open(cntFile); lines = f.readlines(); f.close()
    #挿入場所は、!TEMPERATUREのALLに、'otherData=yes'を追加して挿入。
    si = 0
    i = 0
    while i < len(lines):
        i = pyFistr.getNextHeaderName(lines, i, "!TEMPERATURE", "")
        i += 1
        if i >= len(lines):
            break
        words = pyFistr.deleteSp(lines[i]).split(",")
        if words[0] == "ALL":
            #!TEMPERATUREの行に'otherData=yes'を追加
            lines[i-1] = pyFistr.setValName(lines[i-1], "otherData=yes")
            si = i + 1
            ei = pyFistr.skipNextHeader(lines, i)
            break
    if si == 0:
        print("error: could not find !TEMPERATURE card!")
        return
    #tempLinesを追加
    lines = lines[:si] + tempLines + lines[ei:]
    f = open(cntFile, "w"); f.writelines(lines); f.close()

#
#  remakeFistrCntFileForPress
#------------------------------
def remakeFistrCntFileForPress(mapLines, iniFlag):
    """ 圧力のmapping結果をcntファイルに反映"""
    #追加行を作成
    coeff = float(mapCoeff)
    shift = float(mapShift)
    addLines = []
    for line in mapLines:
        if line[0] != "#":
            words = line.split()
            if len(words) >= 3:
                #words[2] = str(float(words[2]) * coeff + shift)
                press = float(words[2]) * coeff + shift
                #計算初回のみ値を修正
                if iniFlag == 0:
                    #値が「0」の場合、zeroに近い値に修正
                    if press == 0.0:
                        press = 1.0e-36     #zeroに近い値
                words[2] = str(press)
                words[1] = "P" + words[1]
                newLine = ", ".join(words) + "\n"
                addLines.append(newLine)
    #cntFileを修正
    cntFile = solidCaseDir + os.sep + "FistrModel_fsi.cnt"
    f = open(cntFile); lines = f.readlines(); f.close()
    #挿入場所は、最初の!DLOADの該当SGRPに挿入する。
    i = 0
    while i < len(lines):
        i = pyFistr.getNextHeaderName(lines, i, "!DLOAD", "")
        i += 1
        if i >= len(lines):
            mess = "error: '" + sgrpName + "' does not exist in !DLOAD.\n"
            print(mess)
        words = pyFistr.deleteSp(lines[i]).split(",")
        if words[0] == sgrpName:
            si = i - 1
            ei = pyFistr.skipNextHeader(lines, i)
            break
    if si < len(lines):
        #cntFile修正、書き込み
        name = "mappingSP_" + sgrpName + ".map"
        grpid = pyFistr.getValName(lines[si], "GRPID")
        words = ["!DLOAD", "GRPID="+grpid, "forceType=mapping", "mapFile="+name,
                 "mapCoeff="+mapCoeff, "mapShift="+mapShift]
        lines[si] = ", ".join(words) + "\n"
        lines[si+1] = sgrpName + ", S, 0.0\n"
        newLines = lines[:si+2] + addLines + lines[ei:]
        f = open(cntFile, "w"); f.writelines(newLines); f.close()

#
#  runOpenFOAM
#---------------
def runOpenFOAM():
    """ openFOAMの起動"""
    global OF_pid, timePrec
    contDict = pyTreeFoam.case().getControlDict()
    timePrec = contDict["timePrecision"]
    solver = contDict["application"]
    if int(nProcs) > 1:
        #parallel設定
        comm = "mpirun -np " + nProcs + " " + solver + " -parallel 2>&1 | teeCorrect.py solve.log"
    else:
        #singleコア
        comm = solver + " 2>&1 | teeCorrect.py solve.log"
    #OpenFOAMを起動
    procOF = subprocess.Popen(
        comm,
        shell=True)
    OF_pid = procOF.pid     #OpenFOAMのpidを取得
    return procOF

#
#  runFistr
#-----------
def runFistr(sect, endStep):
    """ FrontISTRを起動"""
    saveDir = os.getcwd()
    os.chdir(solidCaseDir)
    logFile = fluidCaseDir + "/solidSolve.log"
    fo = open(logFile, "ab")
    t = time.time()
    comm = solidAnaComm + " -t " + nThreads
    procFistr = subprocess.run(
        comm,
        shell=True,
        stdout=fo,
        stderr=fo
        )
    l = time.time() - t
    l = int(l * 100 + 0.5) / 100
    cont = sect + " -> " + str(endStep) + " steps"
    line = "FrontISTR total time (" + cont + ") ----> " + str(l) + " sec\n\n"
    fo.write(line.encode())
    fo.close()
    os.chdir(saveDir)

#
#  putStreamEachProcs
#---------------------
def putStreamEachProcs(procNo, title, conts, q):
    """ 並列処理用のプロセス。並列数分起動される。
    自身のstreamにcontsを出力する。"""
    putProcStream(procNo, title, conts)
    q.put("")

#
#  addDisplacements
#--------------------
def addDisplacements(chtssDispPtc, fsiDispPtc):
    """ 引数の変位の値を加算して返す"""
    if len(chtssDispPtc) == 0:
        return fsiDispPtc
    if len(fsiDispPtc) == 0:
        return chtssDispPtc
    #両者を加算して戻す
    newLocPtc = []
    chtssDispPtc.sort()
    fsiDispPtc.sort()
    for i in range(len(fsiDispPtc)):
        loc, fsiDisp = fsiDispPtc[i]
        _loc, chtssDisp = chtssDispPtc[i]
        vals = []
        for ii in range(3):
            val = float(fsiDisp[ii]) + float(chtssDisp[ii])
            vals.append(str(val))
        ptc = [loc, vals]
        newLocPtc.append(ptc)
    return newLocPtc

#
#  sendContsToStream
#--------------------
def sendContsToStream(title, conts):
    """ dataをstreamにセットする。"""
    #並列数分のprocessを起動する
    queue = []
    procs = []
    for i in range(int(nProcs)):
        q = multiprocessing.Queue()
        queue.append(q)
        p = multiprocessing.Process(
            target=putStreamEachProcs,
            args=(str(i), title, conts, q))
        p.start()
        procs.append(p)
    #全セットprocessの終了を待つ
    for i in range(int(nProcs)):
        a = queue[i].get()

#
#  readDispFromResFile
def readDispFromResFile(resFile):
    """ 変位データをresFileから取得"""
    #resFileを読み込み
    f = open(resFile);
    #*dataまでskip
    while True:
        line = f.readline()
        if line[:len("*data")] == "*data":
            break
    #node数読み込み
    line = f.readline()
    nNodes = int(line.split()[0])
    #節点の値の位置を取得
    _dummy = f.readline()
    line = f.readline()
    words = line.split()
    nums = list(map(int, words))
    nVals = sum(nums)
    dispLoc = 0
    flag = 0
    for i in range(len(words)):
        valName = f.readline().split()[0]
        if valName == "DISPLACEMENT":
            flag = 1
        else:
            if flag == 0:
                dispLoc += nums[i]
    #DISPLACEMENTの値を読み込む
    dispDict = {}
    n = 0
    while True:
        #nodeNo取得
        line = f.readline()
        nodeNo = int(line.split()[0])
        #disp = [nodeNo]
        #disp読み込み
        vals = []
        while True:
            line = f.readline()
            words = line.split()
            vals += words
            if len(vals) >= nVals:
                break
        dispWords = vals[dispLoc:dispLoc+3]
        dispVal = list(map(float, dispWords))
        dispDict[nodeNo] = dispVal
        n += 1
        if n >= nNodes:
            break
    f.close()
    return dispDict

#
#  createFsiDispPtc
def createFsiDispPtc(nodeDict):
    """ 変位の点群データ取得"""
    #計算したstepを取得
    step = fsiDict["endStep"]
    #resFile名を取得、読み込み
    resFile = solidCaseDir + "/FistrModel.res.0." + str(step)
    dispDict = readDispFromResFile(resFile)
    #点群作成
    dispPtc = []
    for nodeNo in nodeDict.keys():
        disp = dispDict[nodeNo]
        loc = nodeDict[nodeNo]
        dispPtc.append([loc, disp])
    return dispPtc

#
#  createDispPtcFileFromPtc
#-----------------------
def createDispPtcFileFromPtc(locDisp):
    """ 変位の点群dataから転送用のテキストデータを作成"""
    lines = []
    for loc, disp in locDisp:
        data = loc + disp
        line = " ".join(map(str, data)) + "\n"
        lines.append(line)
    lines.append("#end\n")
    #各streamにdataセット
    conts = "".join(lines)
    return conts

#
#  createChtssDispPtc
def createChtssDispPtc(nodeDict):
    """ chtss変位の点群データ取得"""
    #resFile名を取得、読み込み
    resFile = solidCaseDir + "/FistrModel.res.0.1"
    dispDict = readDispFromResFile(resFile)
    #点群作成
    dispPtc = []
    for nodeNo in nodeDict.keys():
        disp = dispDict[nodeNo]
        loc = nodeDict[nodeNo]
        dispPtc.append([loc, disp])
    return dispPtc

#
#  createTempPtcFile
#--------------------
def createTempPtcFile(nodeDict, tempLines):
    """ 温度の点群fileを作成"""
    #温度の点群データを取得作成
    locTemp = createTempPtc(nodeDict, tempLines)
    #点群ファイル作成
    lines = []
    for loc, temp in locTemp:
        data = loc + [temp]
        line = " ".join(map(str, data)) + "\n"
        lines.append(line)
    lines.append("#end\n")
    #各stream用data作成
    conts = "".join(lines)
    return conts

#
#  createTempPtc
def createTempPtc(nodeDict, tempLines):
    """ 温度の点群データ取得"""
    tempDict = {}
    for line in tempLines:
        words = pyFistr.deleteSp(line).split(",")
        nodeNo = int(words[0])
        temp = float(words[1])
        tempDict[nodeNo] = temp
    tempPtc = []
    for nodeNo in nodeDict.keys():
        temp = tempDict[nodeNo]
        loc = nodeDict[nodeNo]
        tempPtc.append([loc, temp])
    return tempPtc

#
#  createRestartFileForFistr
#----------------------------
def createRestartFileForFistr(currTime, runStep):
    """ restartFileを作成する"""
    deltaT = str(float(currTime) - float(beforeTime))
    writeInterval = couplingDict["writeInterval"]
    if pyCoupling.isWriteInterval(currTime, writeInterval, deltaT) == True:
        if float(currTime) != float(couplingDict["startTime"]):
            #restartを保存
            if fsi == "yes":
                endStep = runStep
                restartFile = solidCaseDir + "/FistrModel.restart.0"
                newFile = solidCaseDir + "/FistrModel.restart_" + currTime + "_" + str(endStep) + ".0"
                shutil.copy(restartFile, newFile)
            if cht == "yes":
                endStep =runStep
                restartFile = solidCaseDir + "/FistrModel_cht.restart.0"
                newFile = solidCaseDir + "/FistrModel_cht.restart_" + currTime + "_" + str(endStep) + ".0"
                shutil.copy(restartFile, newFile)

#
#  clearPressPtcResFile
#-----------------------
def clearPressPtcResFile(beforeTime, currTime, skipSP, skipST, runStep):
    """ 圧力の点群ファイル、resファイルを削除"""
    global delTimes, stepTimeDict, delSteps
    delFiles = []
    delFolders = []
    #起動時のstartStepを記録
    if len(stepTimeDict) == 0:
        if fsi == "yes":
            startStep = fsiDict["startStep"]
        else:
            startStep = chtDict["startStep"]
        stepTimeDict[startStep] = beforeTime
    #OpenFOAM側のファイル削除
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data/"
    delsT = delTimes.copy()
    delsS = delSteps.copy()
    for i in range(len(delTimes)):
        delTime = delsT[i]
        if float(delTime) < float(beforeTime):
            deltaT = str(float(currTime) - float(beforeTime))
            isWriteInterval = pyCoupling.isWriteInterval(delTime, writeInterval, deltaT)
            delStep = delsS[i]
            if isWriteInterval == False:
                #削除可能な時間（writeInterval以外）-------
                #圧力の点群ファイル
                pattern = solidCaseDir + "/" + patchName + "_faceCenterPress_" + delTime + "_*.particles"
                fileNames = getAllProcsFiles(pattern, int(nProcs))                
                delFiles += fileNames
                #dispを削除
                pattern = dataDir + patchName + "_disp_" + delStep + "_*tm"
                fileNames = getAllProcsFiles(pattern, int(nProcs))
                delFiles += fileNames
                #facePressを削除
                pattern = dataDir + patchName + "_facePress_" + delTime + "_*tm"
                fileNames = getAllProcsFiles(pattern, int(nProcs))
                delFiles += fileNames
                #tempを削除
                pattern = dataDir + patchName + "_temp_" + delStep + "_*tm"
                fileNames = getAllProcsFiles(pattern, int(nProcs))
                delFiles += fileNames
                #faceHeatFluxを削除
                pattern = dataDir + patchName + "_faceHeatFlux_" + delTime + "_*tm"
                fileNames = getAllProcsFiles(pattern, int(nProcs))
                delFiles += fileNames
            else:
                #writeInterval時の場合------------
                #圧力の点群ファイル
                pattern = solidCaseDir + "/" + patchName + "_faceCenterPress_" + delTime + "_*.particles"
                fileNames = getAllProcsFiles(pattern, int(nProcs))                
                delFiles += fileNames
                #dispをrename
                pattern = dataDir + patchName + "_disp_" + delStep + "_*tm"
                fileNames = getAllProcsFiles(pattern, int(nProcs))
                delete2charaFromEndName(fileNames)
                #facePressをrename
                pattern = dataDir + patchName + "_facePress_" + delTime + "_*tm"
                fileNames = getAllProcsFiles(pattern, int(nProcs))
                delete2charaFromEndName(fileNames)
                #tempをrename
                pattern = dataDir + patchName + "_temp_" + delStep + "_*tm"
                fileNames = getAllProcsFiles(pattern, int(nProcs))
                delete2charaFromEndName(fileNames)
                #faceHeatFluxをrename
                pattern = dataDir + patchName + "_faceHeatFlux_" + delTime + "_*tm"
                fileNames = getAllProcsFiles(pattern, int(nProcs))
                delete2charaFromEndName(fileNames)
            #listの項目を削除
            delTimes.remove(delTime)
            delSteps.remove(delStep)

    #保存するstepを取得
    deltaT = str(float(currTime) - float(beforeTime))
    if pyCoupling.isWriteInterval(currTime, writeInterval, deltaT) == True:
        if float(currTime) != float(startTime):
            endStep = runStep
            stepTimeDict[endStep] = currTime
    #resファイル,フォルダを取得
    #forntIstr側の削除
    if skipSP == 0:
        #fsiの結果を削除
        startStep = fsiDict["beforeStartStep"]
        endStep = fsiDict["startStep"]
        resFiles  = [solidCaseDir + "/FistrModel.res.0." + str(endStep)]
        vtuFiles = [solidCaseDir + "/FistrModel.vis_psf." + ("%04d" % endStep) + ".pvtu"]
        vtuFolders = [solidCaseDir + "/FistrModel.vis_psf." + ("%04d" % endStep)]
        for resFile in resFiles:
            step = int(resFile.split(".")[-1])
            if (startStep <= step and step <= endStep):
                if not (step in stepTimeDict.keys()):
                    delFiles.append(resFile)            #削除
        for i in range(len(vtuFiles)):
            step = int(vtuFiles[i].split(".")[-2])
            if (startStep <= step and step <= endStep):
                if not (step in stepTimeDict.keys()):
                    delFiles.append(vtuFiles[i])        #削除
                    delFolders.append(vtuFolders[i])    #削除
    if skipST == 0:
        #chtの結果を削除
        #startStep = 0
        #endStep = chtDict["startStep"] - chtDict["beforeStartStep"]
        #startStep = chtDict["beforeStartStep"]
        #endStep = chtDict["startStep"]
        endStep = chtDict["endStep"] - chtDict["startStep"]
        resFiles = [solidCaseDir + "/FistrModel_temp.res.0." + str(endStep)]
        vtuFiles = [solidCaseDir + "/FistrModel_cht.vis_psf." + ("%04d" % endStep) + ".pvtu"]
        vtuFolders = [solidCaseDir + "/FistrModel_cht.vis_psf." + ("%04d" % endStep)]
        for resFile in resFiles:
            step = int(resFile.split(".")[-1])
            #if (startStep <= step and step <= endStep):
            #if (startStep >= step and step <= endStep):
            if (step <= endStep):
                if not(step in stepTimeDict.keys()):
                    delFiles.append(resFile)
        for i in range(len(vtuFiles)):
            step = int(vtuFiles[i].split(".")[-2])
            #if (startStep <= step and step <= endStep):
            #if (startStep >= step and step <= endStep):
            if (step <= endStep):
                if not (step in stepTimeDict.keys()):
                    delFiles.append(vtuFiles[i])
                    delFolders.append(vtuFolders[i])
    #file、folderを削除
    for delFile in delFiles:
        if os.path.exists(delFile) == True:
            os.remove(delFile)
    for delFolder in delFolders:
        if os.path.isdir(delFolder) == True:
            shutil.rmtree(delFolder)

#
#  renameTemporaryFile
#----------------------
def renameTemporaryFile():
    """ temporaryのfile名を修正"""
    global delTimes
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data/"
    #削除
    delFiles = []
    for delTime in delTimes[:-1]:
        pattern = dataDir + patchName + "_facePress_" + delTime + "_*tm"
        delFiles = getAllProcsFiles(pattern, int(nProcs))
        delStep = str(int(float(delTime) / float(couplingDict["deltaT"]) + 0.5))
        pattern = dataDir + patchName + "_disp_" + delStep + "_*tm"
        delFiles += getAllProcsFiles(pattern, int(nProcs))
    for fileName in delFiles:
        if os.path.exists(fileName):
            os.remove(fileName)
    #rename
    pattern = dataDir + patchName + "*tm"
    names = glob.glob(pattern)
    for name in names:
        if os.path.isfile(name) == True:
            os.rename(name, name[:-2])

#
#  getAllProcsFiles
def getAllProcsFiles(pattern, np):
    """ 全proc数のfile名を取得して返す"""
    names = pattern.split("*")
    files = []
    for i in range(np):
        fileName = names[0] + str(i) + names[1]
        files.append(fileName)
    return files

#
#  delete2charaFromEndName
def delete2charaFromEndName(fileNames):
    """ fileNameの最後の2文字(tm)を削除"""
    for fileName in fileNames:
        if os.path.exists(fileName):
            os.rename(fileName, fileName[:-2])

#
#  setDeltaTInCnt_fsi
#----------------------
def setDeltaTInCnt_fsi(fsiDict, currTime, deltaT, iniFlag, runStep):
    """ fsiのcntファイルを更新する。"""
    stepTime = float(deltaT)
    #cnt読み込み
    cntFile = solidCaseDir + "/FistrModel_fsi.cnt"
    f = open(cntFile); lines = f.readlines(); f.close()
    #時間増分の設定
    i = pyFistr.getNextHeaderName(lines, 0, "!DYNAMIC", "")
    i += 2
    words = pyFistr.deleteSp(lines[i]).split(",")
    #endStepを取得セット
    if iniFlag == 0:
        #計算初回は、1 step 減らす
        startStep = runStep - 1
        cTime = float(currTime) - stepTime
        orgStep = startStep
    else:
        #通常は、そのまま
        startStep = runStep
        cTime = float(currTime)
        orgStep = 0
    deltaT = stepTime
    writeInterval = float(couplingDict["writeInterval"])
    nSteps = int(couplingDict["nStepsOF"])
    endStep = pyCoupling.getEndStepFromCouplingWriteInterval(
        startStep, cTime, deltaT, nSteps, writeInterval, orgStep)
    words[2] = str(endStep)         #endStep
    #時間増分セット
    words[3] = str(stepTime)        #時間増分
    lines[i] = ", ".join(words) + "\n"
    #restartの設定
    if iniFlag == 0 and fsiDict["runStep"] == 0:
        #初回は、restartさせずに起動
        i = pyFistr.getNextHeaderName(lines, 0, "!RESTART", "")
        lines[i] = pyFistr.setValName(lines[i], "FREQUENCY=1")
    else:
        #restartで起動する様に修正
        i = pyFistr.getNextHeaderName(lines, 0, "!RESTART", "")
        lines[i] = pyFistr.setValName(lines[i], "FREQUENCY=-1")
    #cntFile保存
    f = open(cntFile, "w"); f.writelines(lines); f.close()
    #変数を更新
    fsiDict["corrStep"] = runStep % int(fsiDict["nSteps"])
    fsiDict["runStep"] = startStep
    fsiDict["beforeStartStep"] = fsiDict["startStep"]
    fsiDict["startStep"] = startStep
    fsiDict["endStep"] = endStep
    fsiDict["beforeCurrTime"] = currTime
    return fsiDict

#
#  setDeltaTInCnt_cht
#---------------------
def setDeltaTInCnt_cht(chtDict, currTime, deltaT, iniFlag, runStep):
    """ chtのcntファイルを更新する。"""
    stepTime = float(deltaT)
    #cnt読み込み
    cntFile = solidCaseDir + "/FistrModel_cht.cnt"
    f = open(cntFile); lines = f.readlines(); f.close()
    #時間増分、終了stepを修正
    i = pyFistr.getNextHeaderName(lines, 0, "!HEAT", "")
    i += 1
    words = pyFistr.deleteSp(lines[i]).split(",")
    #時間増分
    if iniFlag == 0:
        #計算初回は、1 step 減らす
        startStep = runStep - 1
        cTime = float(currTime) - stepTime
        orgStep = startStep
    else:
        #通常は、そのまま
        startStep = runStep
        cTime = float(currTime)
        orgStep = 0
    deltaT = stepTime
    writeInterval =float(couplingDict["writeInterval"])
    nSteps = int(couplingDict["nSteps"])
    endStep = pyCoupling.getEndStepFromCouplingWriteInterval(
        startStep, cTime, deltaT, nSteps, writeInterval, orgStep)
    words[0] = str(stepTime)            #時間増分
    endTimeStep = endStep - startStep
    if iniFlag == 0:
        endTimeStep += 1
    endTime = endTimeStep * deltaT
    words[1] = str(endTime)
    lines[i] = ", ".join(words) + "\n"
    #cntFile保存
    f = open(cntFile, "w"); f.writelines(lines); f.close()
    #変数を更新
    chtDict["corrStep"] = runStep % int(chtDict["nSteps"])
    chtDict["runStep"] = startStep
    chtDict["beforeStartStep"] = chtDict["startStep"]
    chtDict["startStep"] = startStep
    chtDict["endStep"] = endStep
    chtDict["beforeCurrTime"] = currTime
    chtDict["endTime"] = endTime
    return chtDict

#
#  roundVal
def roundVal(x, k):
    """ xを有効数字k桁に丸める"""
    x = round(x, k - math.floor(math.log10(abs(x)))- 1)
    return x

#
#  getSubStepsNumInFistr
#-----------------------
def getSubStepsNumInFistr():
    """ FistrModel.cntファイルからsubStep数を取得する。
    !STEPがない場合は、「1」を返す。"""
    nSubSteps = 1
    fileName = solidCaseDir + "/FistrModel.cnt"
    f = open(fileName); lines = f.readlines(); f.close()
    i = pyFistr.getNextHeaderName(lines, 0, "!STEP", "")
    if i < len(lines):
        name = pyFistr.getValName(lines[i], "SUBSTEPS")
        nSubSteps = int(name)
    return nSubSteps

#
#  setReastartFile
#------------------
def setReastartFile(startTime):
    """ FrontISTRのrestartFileを設定する"""
    runStepFsi = 0
    runStepCht = 0
    if fsi == "yes":
        runStepFsi, endStep = setRestartFile_fsi(startTime)
    if cht == "yes":
        runStepCht, endStep = setRestartFile_cht(startTime)
    return runStepFsi, runStepCht

#
#  setRestartFile_fsi
def setRestartFile_fsi(startTime):
    """ fsiの場合:FrontISTRのrestartFileを設定する"""
    pattern = solidCaseDir + "/FistrModel.restart_*.0"
    files = glob.glob(pattern)
    restartFile = ""
    for name in files:
        resTime = name.split("_")[-2]
        if float(resTime) == float(startTime):
            restartFile = name
            break
    if restartFile == "":
        print("error: restart file does not exist. could not restart !!!!\n")
        return
    #runStepを取得
    restFile = ".".join(restartFile.split(".")[:-1])
    runStep = int(restFile.split("_")[-1])
    #runStep = int(restFile.split("_")[-1]) - 1
    #restartFileを設定
    orgName = restFile + ".0"
    newName = solidCaseDir + "/FistrModel.restart.0"
    shutil.copy(orgName, newName)
    #コピーが完了するまで待つ(シェルが認識できるまで待つ)
    time.sleep(1.0)
    comm = ["#!/bin/bash\n"]
    for np in range(int(nProcs)):
        newName = solidCaseDir + "/FistrModel.restart.0"
        comm += ["FILE=" + newName + "\n" ]
        comm += ["while [ ! -e $FILE ]\n"]
        comm += ["do\n"]
        comm += ["  sleep 0.1\n"]
        comm += ["done\n"]
    exeFile = solidCaseDir + "/run"
    f = open(exeFile, "w"); f.writelines(comm); f.close()
    pyFistr.run(solidCaseDir).command("chmod a+x run")
    pyFistr.run(solidCaseDir).command(exeFile)
    #endStepをcntファイルに書き込み
    endStep = runStep + int(nStepsOF)
    #初期の時間増分を設定
    stepTime = float(deltaT)
    cntFile = solidCaseDir + "/FistrModel_fsi.cnt"
    f = open(cntFile); lines = f.readlines(); f.close()
    i = pyFistr.getNextHeaderName(lines, 0, "!DYNAMIC", "")
    i += 2
    words = pyFistr.deleteSp(lines[i]).split(",")
    words[2] = str(endStep)     #endStepを設定
    words[3] = str(stepTime)    #時間増分を設定
    lines[i] = ", ".join(words) + "\n"
    f = open(cntFile, "w"); f.writelines(lines); f.close()
    return runStep, endStep

#
#  setRestartFile_cht
def setRestartFile_cht(startTime):
    """ chtの場合:FrontISTRのrestartFileを設定する"""
    #global endStep
    pattern = solidCaseDir + "/FistrModel_cht.restart_*.0"
    files = glob.glob(pattern)
    restartFile = ""
    for name in files:
        resTime = name.split("_")[-2]
        if float(resTime) == float(startTime):
            restartFile = name
            break
    if restartFile == "":
        print("error: restart file does not exist. could not restart.")
        return
    #startStepを取得
    restFile = ".".join(restartFile.split(".")[:-1])
    runStep = int(restFile.split("_")[-1])
    endStep = runStep + int(nSteps)
    #開始stepの温度結果を取得
    resFile = solidCaseDir + os.sep + "FistrModel_temp.res.0." + str(runStep)
    tempLines = getTempLinesInResFile(resFile)
    #meshFileに温度データを追加
    addTempDataToMeshFile(tempLines)
    #初期の時間増分、終了時間を設定
    couplingInterval = float(deltaT) * int(nStepsOF)
    stepTime = couplingInterval / int(nStepsOF)
    endTime = couplingInterval
    cntFile = solidCaseDir + "/FistrModel_cht.cnt"
    f = open(cntFile); lines = f.readlines(); f.close()
    i = pyFistr.getNextHeaderName(lines, 0, "!HEAT", "")
    i += 1
    words = pyFistr.deleteSp(lines[i]).split(",")
    words[0] = str(stepTime)
    words[1] = str(endTime)
    lines[i] = ", ".join(words) + "\n"
    f = open(cntFile, "w"); f.writelines(lines); f.close()
    return runStep, endStep

#
#  createPidFile
#-----------------
def createPidFile():
    """ signal送信の為、全プロセスのPIDを保存するfile。
    OpenFOAMが起動するPIDから順番に記録し、最後に自身のPIDを記録"""
    np = int(nProcs)
    cont = " " * (np * 10)
    pid = os.getpid()
    word = "%10s" % pid
    cont += word
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/pids"
    f = open(fileName, "w"); f.write(cont); f.close()
    f = open(fileName, "r+b"); mmPid = mmap.mmap(f.fileno(), 0); f.close()
    return mmPid

#
#  createInitialTimesFile
#------------------------
def createInitialTimesFile():
    """ 初期のtimesファイルを作成する。lengthは、200bytesで固定"""
    length = 200
    cont = " " * length
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/times"
    f = open(fileName, "w"); f.write(cont); f.close()
    f = open(fileName, "r+b"); mmTimes = mmap.mmap(f.fileno(), 0); f.close()
    return mmTimes

#
#  createTimeListFile
#---------------------
def createTimeListFile():
    """ 時間のリストを作成"""
    length = 200
    cont = " " * length
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/timeList"
    f = open(fileName, "w"); f.write(cont); f.close()
    f = open(fileName, "r+b"); mmTimeList = mmap.mmap(f.fileno(), 0); f.close()
    return mmTimeList

#
#  createCommonFiles
#-------------------
def createCommonFiles():
    """ data送受信の為の共通fileを設定"""
    mmCommFiles = []
    size = commonSize
    chara = b"\0"
    cont = chara * size
    for np in range(int(nProcs)):
        fileName = fluidCaseDir + "/coupling_FrontISTR/data/common_" + str(np)
        f = open(fileName, "wb"); f.write(cont); f.close()
        f = open(fileName, "r+b"); mmCommFile = mmap.mmap(f.fileno(), 0); f.close()
        mmCommFiles.append(mmCommFile)
    return mmCommFiles

#
#  createCouplingStepFile
#-------------------------
def createCouplingStepFile():
    """ couplingStepファイルの共有。lengthは、200bytesで固定"""
    couplingStepFile = fluidCaseDir + "/coupling_FrontISTR/data/couplingStep"
    length = 200
    cont = " " * length
    f = open(couplingStepFile, "w"); f.write(cont); f.close()
    f = open(couplingStepFile, "r+b"); mmCoupling = mmap.mmap(f.fileno(), 0); f.close()
    return mmCoupling


if __name__ == "__main__":
    import gettext
    gettext.install("app") # replace with the appropriate catalog name
    _ = gettext.gettext

    #初期設定
    deltaT = ""
    startTime = ""
    fluidCaseDir = os.getcwd()
    couplingData = pyCoupling.couplingData(fluidCaseDir)
    couplingDict = couplingData.read()
    startTime = couplingDict["startTime"]
    deltaT = couplingDict["deltaT"]
    writeInterval = couplingDict["writeInterval"]
    nStepsOF = couplingDict["nStepsOF"]
    patchName = couplingDict["mappingPatch"]
    streamBuffer = couplingDict["streamBuffer"]
    maxWaitTime = couplingDict["maxWaitTime"]
    solidCaseDir = couplingDict["solidCaseDir"]
    sgrpName = couplingDict["mappingSgrp"]
    mapCoeff = couplingDict["mappingCoeff"]
    mapShift = couplingDict["mappingShift"]
    nThreads = couplingDict["nThreads"]
    solidAnaComm = couplingDict["solidAnaComm"]
    nSteps = couplingDict["nSteps"]
    nProcs = couplingDict["nProcs"]
    fsi = couplingDict["fsi"]
    cht = couplingDict["cht"]
    chtss = couplingDict["chtss"]
    solidCaseDir = os.path.abspath(solidCaseDir)
    commonSize = int(streamBuffer)
    waitCount = int(maxWaitTime)
    skipSteps = int(nStepsOF)
    nSubSteps = getSubStepsNumInFistr()
    if float(startTime) == 0.0:
        #初期化---------------------------
        #folder内をクリア
        clearDataFolder()
        createCntFile()
        #Fistrのfolder内をクリア
        clearRestartSetting()
        #Fistrの初期設定
        setInitialSettingFistr()
        #計算開始step数
        runStepFsi = 0
        runStepCht = 0
    else:
        #途中から計算開始-----------
        #restartFileとendStepを設定
        runStepFsi, runStepCht = setReastartFile(startTime)
        #テンポラリファイル（tmファイル）を削除
        clearTempFilesInDataFolder()
    #共有メモリの設定
    mmPid = createPidFile()
    mmCommFiles = createCommonFiles()
    mmTimes = createInitialTimesFile()
    mmCoupling = createCouplingStepFile()
    mmTimeList = createTimeListFile()
    #meshのpatch内容を取得(face,nodeの座標を取得)
    createPatchCont()
    #mshのSGRPの内容を取得
    (sgrpDict, nodeDict) = createSgrpCont()
    #openFOAM - FrontISTR coupling
    # beforeTime, currTimeの初期設定"
    currTime = startTime
    beforeTime = str(float(startTime) - float(deltaT))
    #couplig開始
    runCoupling(sgrpDict, nodeDict)

    


