#!/usr/bin/python3
#  coding: utf-8
#
#       patchRun_setHeatFlux_calcTemp.py
#
#       OpenFAOMの熱流速heatFluxをFrontISTRのcntファイルにセットし、
#       温度の点群ファイルを作成する。
#       温度の点群ファイルからOFのpointにmappingした結果fileを
#       作成する。（OF側がこのfileを読み込む）
#
#   23/04/04    新規作成
#      05/04    温度のmappingを修正
#      06/10    setTempPress_getTemp.py → setTempPress_calcTemp.py
#               に変更。計算結果fileを作成
#      06/13    FrontISTRとOpenFOAMが並列で起動する様に、大幅修正。
#

import sys
import os
import time
import mmap
import mappingFromPointCloud as mapping
import pyTreeFoam_ptc as pyTreeFoam
import createPointCloudFromOpenFoam as createPtc
import pyCoupling

waitCount = 5000
waitTime = 0.01

commonSize = 10000


#
#  getMappedFacePointLocs
#---------------------
def getMappedFacePointLocs():
    """ mappingの為の辞書を作成"""
    #mapPointDictを取得
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mapPointDict = {}
    for line in lines:
        words = line.split()
        if len(words) > 4:
            nodeNo = int(words[0])
            if not (nodeNo in mapPointDict.keys()):
                pointLoc = list(map(float, words[1:4]))
                mapPointDict[nodeNo] = pointLoc
    #mapFaceCenterDictを作成
    name = patchName + "_mapFaceCenterNodes_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mapFaceCenterDict = {}
    for line in lines:
        words = line.split()
        if len(words) >= 7:
            faceNo = int(words[0])
            faceLoc = list(map(float, words[1:4]))
            nodeNos = list(map(int, words[4:]))
            mapFaceCenterDict[faceNo] = [faceLoc, nodeNos]
    return mapFaceCenterDict, mapPointDict

#
# createPtcFileFromOF
#----------------------
def createPtcFileFromOF(valueFile, dataDict):
    """ 点群file作成"""
    fluidCaseDir = dataDict["fluidCaseDir"]
    patchName = dataDict["patchName"]
    procNo = dataDict["procNo"]
    #faceCenterLoc読み込み
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    faceCenterFile = dataDir + "/" + patchName + "_faceCenterLocs_" + procNo
    f = open(faceCenterFile); lines = f.readlines(); f.close()
    #faceValueを読み込み
    fp = open(valueFile); pLines = fp.readlines(); fp.close()
    newLines = []
    for i in range(len(lines)):
        newLine = lines[i][:-1] + " " + pLines[i]
        newLines.append(newLine)
    newLines.append("#end\n")
    #streamにセットし、全体制御側に送信
    conts = "".join(newLines)
    return conts

#
#  saveMappingTemp
#-------------------
def saveMappingTemp(mappingDict):
    """ mappingDispの結果を保存。このファイルをOF側が読み込む"""
    saveDir = fluidCaseDir + "/coupling_FrontISTR/data"
    faceNos = list(mappingDict.keys())
    faceNos.sort()
    lines = []
    for faceNo in faceNos:
        temp = mappingDict[faceNo]
        line = str(temp) + "\n"
        lines.append(line)
    #保存のfilename
    timeIdx = int(couplingDataDict["timeIndex"])
    # startStep = timeIdx
    # currTime = float(couplingDataDict["timeName"])
    if (float(beforeTime) == float(couplingDataDict["startTime"])):
        #計算初回
        startStep = timeIdx - 1
        currTime = float(couplingDataDict["startTime"])
        orgStep = startStep
    else:
        startStep = timeIdx
        currTime = float(couplingDataDict["timeName"])
        orgStep = 0
    #deltaT = currTime - float(couplingDataDict["beforeTime"])
    deltaT = float(couplingDataDict["deltaT"])
    nSteps = int(couplingDataDict["nSteps"])
    writeInterval = float(couplingDataDict["writeInterval"])
    endStep = pyCoupling.getEndStepFromCouplingWriteInterval(
        startStep, currTime, deltaT, nSteps, writeInterval, orgStep)
    #endTime = endStep * float(couplingDataDict["deltaT"])
    endTime = currTime + (endStep - startStep) * deltaT
    if procNo == "0":
        print("(py) temperature file was created for OpenFOAM at " + str(endTime) + ".")
    #name = patchName + "_temp_" + timeName + "_" + procNo + "tm"
    name = patchName + "_temp_" + str(endStep) + "_" + procNo + "tm"
    tempFile = saveDir + "/" + name
    f = open(tempFile, "w"); f.writelines(lines); f.close()
    return

#
#  getLocTemperature
#--------------------
def getLocTemperature(lines):
    """ 温度の点群ファイルを取得"""
    locTemp = []
    for line in lines:
        if line[0] != "#":
            words = line.split()
            loc = list(map(float, words[:3]))
            temp = float(words[3])
            locTemp.append([loc, temp])
    return locTemp

#-----------------------
#  setHeatFlux_getTemp
#-----------------------
def setHeatFlux_getTemp(dataDict):
    """ 熱流束をmappingしてfistrの結果を待ち、温度を取得する。"""
    solidCaseDir = dataDict["solidCaseDir"]
    fluidCaseDir = dataDict["fluidCaseDir"]
    patchName = dataDict["patchName"]
    beforeTime = dataDict["beforeTime"]
    timeName = dataDict["timeName"]
    procNo = dataDict["procNo"]
    mesh = mapping.fistrMesh(solidCaseDir)
    #温度の点群ファイル作成
    ptcDir = fluidCaseDir + "/coupling_FrontISTR/data"
    htFile = ptcDir + "/" + patchName + "_faceHeatFlux_" + timeName + "_" + procNo + "tm"
    htConts = createPtcFileFromOF(htFile, dataDict)
    #heatFluxを転送
    title = "heatFlux"
    streamToParent.put(title, htConts)
    #mappingするfaceCenterの座標を予め取得しておく（待ち時間を利用）
    mapFaceCenterDict, mapPointDict = getMappedFacePointLocs()
    #温度の点群dataをstreamから読み込み
    lines = streamToParent.getLines("temperature")
    locTemp = getLocTemperature(lines)
    mapStat, mappingTriDict = mesh.getMappingTriLocs(locTemp, mapFaceCenterDict, mapPointDict)
    #温度をmapping
    mappingDict = mesh.setMappingDataFromTriLocs(locTemp, mapFaceCenterDict, mappingTriDict)
    #mapping結果を保存(このfileをOpenFOAM側が読み込む)
    saveMappingTemp(mappingDict)
    return ""

#
#  beginning
#------------
def beginning(dataDict):
    """ beforeTime, currTimeをファイルに保存"""
    beforeTime = dataDict["beforeTime"]
    timeName = dataDict["timeName"]
    procNo = dataDict["procNo"]
    timeIndex = dataDict["timeIndex"]
    #processor0のみ
    if procNo == "0":
        #時間を書き込み
        couplePress = "9"
        coupleTemp = "0"
        coupleOther = "9"
        cont = pyCoupling.createTimeCont(
            beforeTime, timeName, timeIndex,
            couplePress, coupleTemp, coupleOther)
        mmTimes.seek(0)
        mmTimes.write(cont.encode())
        mmTimes.flush()
    #couplingFileに反映されるまで待つ
    waitUntilCouplingWrote(beforeTime)
    return

#
#  waitUntilCouplingWrote
def waitUntilCouplingWrote(beforeTime):
    """ couplingStepにbeforeTimeが反映されるまで待つ"""
    count = waitCount
    sleepTime = waitTime
    while True:
        mmCoupling.seek(0)
        cont = mmCoupling.read().decode()
        lines = cont.split("\n")
        if len(lines) >= 3:
            bTime = lines[0].split()[1]
            if float(bTime) == float(beforeTime):
                #beforeTimeが反映された場合戻る
                break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  waiteUntilHeatFluxFile
#-------------------------
def waiteUntilHeatFluxFile():
    """ heatFluxFileが出来上がるまで待つ"""
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    fileName = dataDir + "/" + patchName + "_faceHeatFlux_" + timeName + "_" + procNo + "tm"
    #fileが出来上がるまで待つ
    count = waitCount
    sleepTime = waitTime
    while True:
        if os.path.exists(fileName) == True:
            time.sleep(sleepTime)
            break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  createMmapFile
#-----------------
def createMmapFile(fileName):
    """ mmapFileを作成する"""
    f = open(fileName, "r+b"); mmFile = mmap.mmap(f.fileno(), 0); f.close()
    return mmFile



if __name__ == "__main__":
    import gettext
    gettext.install("app") # replace with the appropriate catalog name
    _ = gettext.gettext

    timeName = sys.argv[1]      #currTime
    beforeTime = sys.argv[2]    #前回のtime
    patchName = sys.argv[3]     #patchName
    procNo = sys.argv[4]        #procNo
    nProcs = sys.argv[5]        #nProcs
    timeIndex = sys.argv[6]     #timeIndex
    fluidCaseDir = os.getcwd()
    couplingData = pyCoupling.couplingData(fluidCaseDir)
    couplingDataDict = couplingData.read()
    couplingDataDict["fluidCaseDir"] = fluidCaseDir
    couplingDataDict["timeName"] = timeName
    couplingDataDict["beforeTime"] = beforeTime
    couplingDataDict["patchName"] = patchName
    couplingDataDict["procNo"] = procNo
    couplingDataDict["nProcs"] = nProcs
    couplingDataDict["timeIndex"] = timeIndex
    waitCount = int(couplingDataDict["maxWaitTime"])
    #共有メモリの設定
    commonSize = int(couplingDataDict["streamBuffer"])
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    mmCommFile = createMmapFile(dataDir + "/common_" + procNo)
    mmTimes = createMmapFile(dataDir + "/times")
    mmCoupling = createMmapFile(dataDir + "/couplingStep")
    mmTimeList = createMmapFile(dataDir + "/timeList")
    #streamを定義
    streamToParent = pyCoupling.stream(mmCommFile, commonSize,
                                       waitCount, waitTime )
    beginning(couplingDataDict)
    #終了？
    if (float(timeName) >= float(couplingDataDict["endTime"])):
        exit()
    waiteUntilHeatFluxFile()
    #連続計算
    solidCaseDir = os.path.abspath(couplingDataDict["solidCaseDir"])
    stat = setHeatFlux_getTemp(couplingDataDict)
    if stat != "":
        raise NameError(stat)

    