#!/usr/bin/python3
#  coding: utf-8
#
#       patchRun_setPressHt_calcTempDispChtss.py
#
#       OpenFAOMの熱流速heatFluxをFrontISTRのcntファイルにセットし、
#       温度の点群ファイルを作成する。
#       温度の点群ファイルからOFのpointにmappingした結果fileを
#       作成する。（OF側がこのfileを読み込む）
#
#   23/04/04    新規作成
#      05/04    温度のmappingを修正
#      06/10    setTempPress_getTemp.py → setTempPress_calcTemp.py
#               に変更。計算結果fileを作成
#      06/13    FrontISTRとOpenFOAMの並列起動の為、大幅修正。
#

import sys
import os
import time
import mmap
import mappingFromPointCloud as mapping
import pyCoupling

waitCount = 5000
waitTime = 0.01

commonSize = 10000

#
#  getMappedPointLocs
#---------------------
def getMappedPointLocs():
    """ mappingするpointの座標を取得する"""
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mapPointFaceDict = {}
    mapPointDict = {}
    for line in lines:
        words = line.split()
        if len(words) > 4:
            nodeNo = int(words[0])
            pointLoc = list(map(float, words[1:4]))
            faceNodes = list(map(int, words[4:]))
            mapPointFaceDict[nodeNo] = [pointLoc, faceNodes]
            mapPointDict[nodeNo] = pointLoc
    return mapPointFaceDict, mapPointDict

#
#  getMappedFacePointLocs
#---------------------
def getMappedFacePointLocs():
    """ mappingの為の辞書を作成"""
    #mapPointDictを取得
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mapPointDict = {}
    for line in lines:
        words = line.split()
        if len(words) > 4:
            nodeNo = int(words[0])
            if not (nodeNo in mapPointDict.keys()):
                pointLoc = list(map(float, words[1:4]))
                mapPointDict[nodeNo] = pointLoc
    #mapFaceCenterDictを作成
    name = patchName + "_mapFaceCenterNodes_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mapFaceCenterDict = {}
    for line in lines:
        words = line.split()
        if len(words) >= 7:
            faceNo = int(words[0])
            faceLoc = list(map(float, words[1:4]))
            nodeNos = list(map(int, words[4:]))
            mapFaceCenterDict[faceNo] = [faceLoc, nodeNos]
    return mapFaceCenterDict, mapPointDict

#
# createPtcFileFromOF
#----------------------
def createPtcFileFromOF(valueFile, dataDict):
    """ 点群file作成"""
    fluidCaseDir = dataDict["fluidCaseDir"]
    patchName = dataDict["patchName"]
    procNo = dataDict["procNo"]
    #faceCenterLoc読み込み
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    faceCenterFile = dataDir + "/" + patchName + "_faceCenterLocs_" + procNo
    f = open(faceCenterFile); lines = f.readlines(); f.close()
    #faceValueを読み込み
    fp = open(valueFile); pLines = fp.readlines(); fp.close()
    newLines = []
    for i in range(len(lines)):
        newLine = lines[i][:-1] + " " + pLines[i]
        newLines.append(newLine)
    newLines.append("#end\n")
    #streamにセットし、全体制御側に送信
    conts = "".join(newLines)
    return conts

#
#  saveMappingDisp
#------------------
def saveMappingDisp(mappingDict, steps):
    """ mappingDispの結果を保存。このファイルをOF側が読み込む"""
    saveDir = fluidCaseDir + "/coupling_FrontISTR/data"
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = saveDir + "/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    #保存のfilename
    endStep = steps
    #endTime = endStep * float(couplingDataDict["deltaT"])
    if procNo == "0":
        print("(py) displacement file is creating for OpenFOAM at " + str(endStep) + "...", flush=True)
    name = patchName + "_disp_" + str(endStep) + "_" + procNo + "tm"
    dispFile = saveDir + "/" + name
    f = open(dispFile, "w")
    for line in lines:
        words = line.split()
        nodeNo = int(words[0])
        disp = mappingDict[nodeNo]
        line = " ".join(list(map(str, disp))) + "\n"
        f.write(line)
    f.close()
    return

#
#  getLocDisplacement
#---------------------
def getLocDisplacement():
    """ 変位の点群ファイルStreamから読み込み"""
    #straemから読み込み
    lines, steps = streamToParent.getLinesAtSteps("displacement")
    locDisp = []
    for line in lines:
        if line[0] != "#":
            words = line.split()
            loc = list(map(float, words[:3]))
            disp = list(map(float, words[3:]))
            locDisp.append([loc, disp])
    return locDisp, steps

#
#  saveMappingTemp
#-------------------
def saveMappingTemp(mappingDict):
    """ mappingDispの結果を保存。このファイルをOF側が読み込む"""
    saveDir = fluidCaseDir + "/coupling_FrontISTR/data"
    faceNos = list(mappingDict.keys())
    faceNos.sort()
    lines = []
    for faceNo in faceNos:
        temp = mappingDict[faceNo]
        line = str(temp) + "\n"
        lines.append(line)
    #保存のfilename
    timeIdx = int(couplingDataDict["timeIndex"])
    if (float(beforeTime) == float(couplingDataDict["startTime"])):
        #計算初回
        startStep = timeIdx - 1
        currTime = float(couplingDataDict["startTime"])
        orgStep = startStep
    else:
        startStep = timeIdx
        currTime = float(couplingDataDict["timeName"])
        orgStep = 0
    deltaT = currTime - float(couplingDataDict["beforeTime"])
    nSteps = int(couplingDataDict["nSteps"])
    writeInterval = float(couplingDataDict["writeInterval"])
    endStep = pyCoupling.getEndStepFromCouplingWriteInterval(
        startStep, currTime, deltaT, nSteps, writeInterval, orgStep)
    endTime = endStep * float(couplingDataDict["deltaT"])
    if procNo == "0":
        print("(py) temperature file is creating for OpenFOAM at " + str(endTime) + ".", flush=True)
    name = patchName + "_temp_" + str(endStep) + "_" + procNo + "tm"
    tempFile = saveDir + "/" + name
    f = open(tempFile, "w"); f.writelines(lines); f.close()
    return

#
#  getLocTemperature
#--------------------
def getLocTemperature(lines):
    """ 温度の点群ファイルを取得"""
    locTemp = []
    for line in lines:
        if line[0] != "#":
            words = line.split()
            loc = list(map(float, words[:3]))
            temp = float(words[3])
            locTemp.append([loc, temp])
    return locTemp

#
#  isMappingTriFile
def isMappingTriFile():
    """ mapping用の3点を取得したfileが存在するかチェック"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    return os.path.exists(fileName)

#
#  createMappingTriFile
def createMappingTriFile(mappingTriDict):
    """ mapping用の3点の座標をfileに保存する"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    lines = []
    for key in mappingTriDict.keys():
        faceNo = key
        word = str(faceNo)
        triLocs = mappingTriDict[key]
        loc0, loc1, loc2 = triLocs
        word0 = " ".join(list(map(str, loc0)))
        word1 = " ".join(list(map(str, loc1)))
        word2 = " ".join(list(map(str, loc2)))
        line = " ".join([word, word0, word1, word2]) + "\n"
        lines.append(line)
    f = open(fileName, "w"); f.writelines(lines); f.close()

#
#  readMappingTriFile
def readMappingTriFile():
    """ 保存したmapping用の3点の座標を読み込む"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mappingTriDict = {}
    for line in lines:
        words = line.split()
        faceNo = int(words[0])
        key = faceNo
        loc0 = list(map(float, words[1:4]))
        loc1 = list(map(float, words[4:7]))
        loc2 = list(map(float, words[7:10]))
        mappingTriDict[key] = [loc0, loc1, loc2]
    return mappingTriDict


#-----------------------
#  setHeatFlux_getTemp
#-----------------------
def setHeatFlux_getTemp(dataDict):
    """ 熱流束をmappingしてfistrの結果を待ち、温度を取得する。"""
    solidCaseDir = dataDict["solidCaseDir"]
    fluidCaseDir = dataDict["fluidCaseDir"]
    patchName = dataDict["patchName"]
    beforeTime = dataDict["beforeTime"]
    timeName = dataDict["timeName"]
    procNo = dataDict["procNo"]
    mesh = mapping.fistrMesh(solidCaseDir)
    
    #圧力の点群ファイル作成
    ptcDir = fluidCaseDir + "/coupling_FrontISTR/data"
    pressFile = ptcDir + "/" + patchName + "_facePress_" + timeName + "_" + procNo + "tm"
    pressConts = createPtcFileFromOF(pressFile, dataDict)
    #圧力の点群データを送信 
    title = "pressurePtcData"
    streamToParent.put(title, pressConts)
    #圧力の点群file保存
    pressPtcFile = solidCaseDir + "/" + patchName + "_faceCenterPress_" + timeName + "_" + procNo + ".particles"
    f = open(pressPtcFile, "w"); f.write(pressConts); f.close()

    streamToParent.waitUntilBufferEmpty()
    #温度の点群ファイル作成
    htFile = ptcDir + "/" + patchName + "_faceHeatFlux_" + timeName + "_" + procNo + "tm"
    htConts = createPtcFileFromOF(htFile, dataDict)
    #heatFluxを転送
    title = "heatFlux"
    streamToParent.put(title, htConts)

    #変位を取得
    #mappingするpointの座標を予め取得しておく（待ち時間を利用）
    mapPointFaceDict, mapPointDict = getMappedPointLocs()
    #変位の点群dataをstreamから読み込み
    locDisp, steps = getLocDisplacement()
    #mappingTriFile(mapping用の3点の座標を保存)があるか？
    if isMappingTriFile() == False:    
        #存在しない場合、mappingTriDictを取得
        mapStat, mappingTriDict = mesh.getMappingTriLocs(locDisp, mapPointFaceDict, mapPointDict)
        if mapStat != "OK":
            return "error in displacement mapping."
        #mappingTriFileを作成
        createMappingTriFile(mappingTriDict)
    else:
        #fileを読み込む
        mappingTriDict = readMappingTriFile()
    #変位をmapping
    mappingDict = mesh.setMappingDataFromTriLocs(locDisp, mapPointFaceDict, mappingTriDict)
    #mapping結果を保存(このfileをOpenFOAM側が読み込む)
    saveMappingDisp(mappingDict, steps)

    #温度を取得
    #mappingするfaceCenterの座標を予め取得しておく（待ち時間を利用）
    mapFaceCenterDict, mapPointDict = getMappedFacePointLocs()
    #変位の点群dataをstreamから読み込み
    lines = streamToParent.getLines("temperature")
    locTemp = getLocTemperature(lines)
    mapStat, mappingTriDict = mesh.getMappingTriLocs(locTemp, mapFaceCenterDict, mapPointDict)
    #温度をmapping
    mappingDict = mesh.setMappingDataFromTriLocs(locTemp, mapFaceCenterDict, mappingTriDict)
    #mapping結果を保存(このfileをOpenFOAM側が読み込む)
    saveMappingTemp(mappingDict)
    return ""

#
#  beginning
#------------
def beginning(dataDict):
    """ beforeTime, currTimeをファイルに保存"""
    beforeTime = dataDict["beforeTime"]
    timeName = dataDict["timeName"]
    procNo = dataDict["procNo"]
    timeIndex = dataDict["timeIndex"]
    #processor0のみ
    if procNo == "0":
        #時間を書き込み
        couplePress = "0"
        coupleTemp = "0"
        coupleOther = "9"
        cont = pyCoupling.createTimeCont(
            beforeTime, timeName, timeIndex,
            couplePress, coupleTemp, coupleOther)
        mmTimes.seek(0)
        mmTimes.write(cont.encode())
        mmTimes.flush()
    #couplingFileに反映されるまで待つ
    waitUntilCouplingWrote(beforeTime)
    return

#
#  waitUntilCouplingWrote
def waitUntilCouplingWrote(beforeTime):
    """ couplingStepにbeforeTimeが反映されるまで待つ"""
    count = waitCount
    sleepTime = waitTime
    while True:
        mmCoupling.seek(0)
        cont = mmCoupling.read().decode()
        lines = cont.split("\n")
        if len(lines) >= 3:
            bTime = lines[0].split()[1]
            if float(bTime) == float(beforeTime):
                #beforeTimeが反映された場合戻る
                break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  waitUntilPressFile
#---------------------
def waitUntilPressFile():
    """ pressFileが出来上がるまで待つ"""
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    name = patchName + "_facePress_" + timeName + "_" + procNo + "tm"
    fileName = dataDir + "/" + name
    #fileが出来上がるまで待つ
    count = waitCount
    sleepTime = waitTime
    while True:
        if os.path.exists(fileName) == True:
            time.sleep(sleepTime)
            break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  waitUntilHeatFluxFile
#-------------------------
def waitUntilHeatFluxFile():
    """ heatFluxFileが出来上がるまで待つ"""
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    fileName = dataDir + "/" + patchName + "_faceHeatFlux_" + timeName + "_" + procNo + "tm"
    #fileが出来上がるまで待つ
    count = waitCount
    sleepTime = waitTime
    while True:
        if os.path.exists(fileName) == True:
            time.sleep(sleepTime)
            break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  createMmapFile
#-----------------
def createMmapFile(fileName):
    """ mmapFileを作成する"""
    f = open(fileName, "r+b"); mmFile = mmap.mmap(f.fileno(), 0); f.close()
    return mmFile



if __name__ == "__main__":
    import gettext
    gettext.install("app") # replace with the appropriate catalog name
    _ = gettext.gettext

    timeName = sys.argv[1]      #currTime
    beforeTime = sys.argv[2]    #前回のtime
    patchName = sys.argv[3]     #patchName
    procNo = sys.argv[4]        #procNo
    nProcs = sys.argv[5]        #nProcs
    timeIndex = sys.argv[6]     #timeIndex
    fluidCaseDir = os.getcwd()
    couplingData = pyCoupling.couplingData(fluidCaseDir)
    couplingDataDict = couplingData.read()
    couplingDataDict["fluidCaseDir"] = fluidCaseDir
    couplingDataDict["timeName"] = timeName
    couplingDataDict["beforeTime"] = beforeTime
    couplingDataDict["patchName"] = patchName
    couplingDataDict["procNo"] = procNo
    couplingDataDict["nProcs"] = nProcs
    couplingDataDict["timeIndex"] = timeIndex
    waitCount = int(couplingDataDict["maxWaitTime"])
    #共有メモリの設定
    commonSize = int(couplingDataDict["streamBuffer"])
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    mmCommFile = createMmapFile(dataDir + "/common_" + procNo)
    mmTimes = createMmapFile(dataDir + "/times")
    mmCoupling = createMmapFile(dataDir + "/couplingStep")
    mmTimeList = createMmapFile(dataDir + "/timeList")
    #streamを定義
    streamToParent = pyCoupling.stream(mmCommFile, commonSize,
                                       waitCount, waitTime )
    beginning(couplingDataDict)
    #終了？
    if (float(timeName) >= float(couplingDataDict["endTime"])):
        exit()
    waitUntilPressFile()
    waitUntilHeatFluxFile()
    #couplingの状況を読み込み
    mmCoupleDict = pyCoupling.getTimeCont(mmCoupling)
    #連続計算
    solidCaseDir = os.path.abspath(couplingDataDict["solidCaseDir"])
    stat = setHeatFlux_getTemp(couplingDataDict)
    if stat != "":
        raise NameError(stat)

    