#!/usr/bin/python3
#  coding: utf-8
#
#       patchRun_setPress_calcDisp.py
#
#       OpenFOAMの圧力pをFrontISTRのcntファイルにセットし、
#       変位の点群ファイルを作成する。
# 　　　変位の点群ファイルからOFのpointにmappingした結果fileを
#       作成する。（OF側がこのfileを読み込む）
#
#   23/01/24    新規作成
#      04/05    couplingの判断をbeginning後に変更
#      05/17    isCoupling:fsi,chtssのcouplingを修正
#      06/04    scalarFieldからの起動で計算するように修正
#               （pressを出力直後に起動する）
#               従来は、pointDisplacement fieldから起動
#      06/13    FrontISTRとOpenFOAMを並列で起動する様に、大幅修正
#      06/20    並列処理のバグ修正
#


import sys
import os
import time
import mmap
import mappingFromPointCloud as mapping
import pyCoupling

waitCount = 5000
waitTime = 0.01

commonSize = 10000



#
#  getMappedPointLocs
#---------------------
def getMappedPointLocs():
    """ mappingするpointの座標を取得する"""
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mapPointFaceDict = {}
    mapPointDict = {}
    for line in lines:
        words = line.split()
        if len(words) > 4:
            nodeNo = int(words[0])
            pointLoc = list(map(float, words[1:4]))
            faceNodes = list(map(int, words[4:]))
            mapPointFaceDict[nodeNo] = [pointLoc, faceNodes]
            mapPointDict[nodeNo] = pointLoc
    return mapPointFaceDict, mapPointDict

#
# createPtcFile
#---------------
def createPtcFile(pressFile):
    """ 圧力の点群file作成"""
    global mmList
    #faceCenterLoc読み込み
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    faceCenterFile = dataDir + "/" + patchName + "_faceCenterLocs_" + procNo
    f = open(faceCenterFile); lines = f.readlines(); f.close()
    #facePressを読み込み
    fp = open(pressFile); pLines = fp.readlines(); fp.close()
    newLines = []
    for i in range(len(lines)):
        newLine = lines[i][:-1] + " " + pLines[i]
        newLines.append(newLine)
    newLines.append("#end\n")
    #streamにセットし、送信
    cont = "".join(newLines)
    title = "pressurePtcData"
    streamToParent.put(title, cont)
    #圧力のptcファイルを保存
    name = patchName + "_faceCenterPress_" + timeName + "_" + procNo + ".particles"
    ptcFileName = solidCaseDir + "/" + name
    f = open(ptcFileName, "w"); f.writelines(newLines); f.close()
    return

#
#  saveMappingDisp
#------------------
def saveMappingDisp(mappingDict, steps):
    """ mappingDispの結果を保存。このファイルをOF側が読み込む"""
    saveDir = fluidCaseDir + "/coupling_FrontISTR/data"
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = saveDir + "/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    #保存のfilename
    endStep = steps
    #endTime = endStep * float(couplingDict["deltaT"])
    if procNo == "0":
        print("(py) displacement file is creating for OpenFOAM at " + str(endStep) + "...")
    name = patchName + "_disp_" + str(endStep) + "_" + procNo + "tm"
    dispFile = saveDir + "/" + name
    f = open(dispFile, "w")
    for line in lines:
        words = line.split()
        nodeNo = int(words[0])
        disp = mappingDict[nodeNo]
        line = " ".join(list(map(str, disp))) + "\n"
        f.write(line)
    f.close()
    return

#
#  getLocDisplacement
#---------------------
def getLocDisplacement():
    """ 変位の点群ファイルStreamから読み込み"""
    #straemから読み込み
    lines, steps = streamToParent.getLinesAtSteps("displacement")
    locDisp = []
    for line in lines:
        if line[0] != "#":
            words = line.split()
            loc = list(map(float, words[:3]))
            disp = list(map(float, words[3:]))
            locDisp.append([loc, disp])
    return locDisp, steps

#
#  isMappingTriFile
def isMappingTriFile():
    """ mapping用の3点を取得したfileが存在するかチェック"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    return os.path.exists(fileName)

#
#  createMappingTriFile
def createMappingTriFile(mappingTriDict):
    """ mapping用の3点の座標をfileに保存する"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    lines = []
    for key in mappingTriDict.keys():
        faceNo = key
        word = str(faceNo)
        triLocs = mappingTriDict[key]
        loc0, loc1, loc2 = triLocs
        word0 = " ".join(list(map(str, loc0)))
        word1 = " ".join(list(map(str, loc1)))
        word2 = " ".join(list(map(str, loc2)))
        line = " ".join([word, word0, word1, word2]) + "\n"
        lines.append(line)
    f = open(fileName, "w"); f.writelines(lines); f.close()

#
#  readMappingTriFile
def readMappingTriFile():
    """ 保存したmapping用の3点の座標を読み込む"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mappingTriDict = {}
    for line in lines:
        words = line.split()
        faceNo = int(words[0])
        key = faceNo
        loc0 = list(map(float, words[1:4]))
        loc1 = list(map(float, words[4:7]))
        loc2 = list(map(float, words[7:10]))
        mappingTriDict[key] = [loc0, loc1, loc2]
    return mappingTriDict

#---------------------
#  setPress_getDisp
#---------------------
def setPress_getDisp():
    """ pressをmappingしてfistrの結果を待ち、変位を取得する"""
    mesh = mapping.fistrMesh(solidCaseDir)
    #ptcDataの読み込み
    ptcDir = fluidCaseDir + "/coupling_FrontISTR/data"
    if float(beforeTime) == float(couplingDict["startTime"]):
        pressFile = ptcDir + "/" + patchName + "_facePress_" + beforeTime + "_" + procNo + "tm"
    else: 
        pressFile = ptcDir + "/" + patchName + "_facePress_" + timeName + "_" + procNo + "tm"
    waitUntilPressFile()
    #圧力の点群ファイルを作成し、送信 
    createPtcFile(pressFile)
    #mappingするpointの座標を予め取得しておく（待ち時間を利用）
    mapPointFaceDict, mapPointDict = getMappedPointLocs()
    #変位の点群dataをstreamから読み込み
    locDisp, steps = getLocDisplacement()
    #mappingTriFile(mapping用の3点の座標を保存)があるか？
    if isMappingTriFile() == False:    
        #存在しない場合、mappingTriDictを取得
        mapStat, mappingTriDict = mesh.getMappingTriLocs(locDisp, mapPointFaceDict, mapPointDict)
        if mapStat != "OK":
            return "error in displacement mapping."
        #mappingTriFileを作成
        createMappingTriFile(mappingTriDict)
    else:
        #fileを読み込む
        mappingTriDict = readMappingTriFile()
    #変位をmapping
    mappingDict = mesh.setMappingDataFromTriLocs(locDisp, mapPointFaceDict, mappingTriDict)
    #mapping結果を保存(このfileをOpenFOAM側が読み込む)
    saveMappingDisp(mappingDict, steps)
    return ""

#
#  beginning
#------------
def beginning(dataDict):
    """ beforeTime, currTimeをファイルに保存"""
    beforeTime = dataDict["beforeTime"]
    timeName = dataDict["timeName"]
    procNo = dataDict["procNo"]
    timeIndex = dataDict["timeIndex"]
    #processor0のみ
    if procNo == "0":
        #時間を書き込み
        couplePress = "0"
        coupleTemp = "9"
        coupleOther = "9"
        cont = pyCoupling.createTimeCont(
            beforeTime, timeName, timeIndex,
            couplePress, coupleTemp, coupleOther)
        mmTimes.seek(0)
        mmTimes.write(cont.encode())
        mmTimes.flush()
    #couplingFileに反映されるまで待つ
    waitUntilCouplingWritten(beforeTime)
    return

#
#  waitUntilCouplingWritten
def waitUntilCouplingWritten(beforeTime):
    """ couplingStepにbeforeTimeが反映されるまで待つ"""
    count = waitCount
    sleepTime = waitTime
    while True:
        mmCoupling.seek(0)
        cont = mmCoupling.read().decode()
        lines = cont.split("\n")
        if len(lines) >= 3:
            bTime = lines[0].split()[1]
            if float(bTime) == float(beforeTime):
                #beforeTimeが反映された場合戻る
                break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  createMmapFile
#-----------------
def createMmapFile(fileName):
    """ mmapFileを作成する"""
    f = open(fileName, "r+b"); mmFile = mmap.mmap(f.fileno(), 0); f.close()
    return mmFile

#
#  deleteMappingTriFile
#-----------------------
def deleteMappingTriFile():
    """ 初回起動時に、mappingTriFileを削除する"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    if os.path.exists(fileName) == True:
        os.remove(fileName)

#
#  waitUntilPressFile
#---------------------
def waitUntilPressFile():
    """ pressFileが出来上がるまで待つ"""
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    name = patchName + "_facePress_" + timeName + "_" + procNo + "tm"
    fileName = dataDir + "/" + name
    #fileが出来上がるまで待つ
    count = waitCount
    sleepTime = waitTime
    while True:
        if os.path.exists(fileName) == True:
            time.sleep(sleepTime)
            break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return


if __name__ == "__main__":
    import gettext
    gettext.install("app") # replace with the appropriate catalog name
    _ = gettext.gettext

    timeName = sys.argv[1]      #currTime
    beforeTime = sys.argv[2]    #前回のtime
    patchName = sys.argv[3]     #patchName
    procNo = sys.argv[4]        #procNo
    nProcs = sys.argv[5]        #nProcs
    timeIndex = sys.argv[6]     #timeIndex
    fluidCaseDir = os.getcwd()
    couplingData = pyCoupling.couplingData(fluidCaseDir)
    couplingDict = couplingData.read()
    couplingDict["fluidCaseDir"] = fluidCaseDir
    couplingDict["timeName"] = timeName
    couplingDict["beforeTime"] = beforeTime
    couplingDict["patchName"] = patchName
    couplingDict["procNo"] = procNo
    couplingDict["nProcs"] = nProcs
    couplingDict["timeIndex"] = timeIndex
    startTime = couplingDict["startTime"]
    writeInterval = couplingDict["writeInterval"]
    nStepsOF = couplingDict["nStepsOF"]
    streamBuffer = couplingDict["streamBuffer"]
    maxWaitTime = couplingDict["maxWaitTime"]
    solidCaseDir = couplingDict["solidCaseDir"]
    sgrpName = couplingDict["mappingSgrp"]
    mapCoeff = couplingDict["mappingCoeff"]
    solidAnaComm = couplingDict["solidAnaComm"]
    nSteps = couplingDict["nSteps"]
    waitCount = int(maxWaitTime)
    #共有メモリの設定
    commonSize = int(streamBuffer)
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    mmCommFile = createMmapFile(dataDir + "/common_" + procNo)
    mmTimes = createMmapFile(dataDir + "/times")
    mmCoupling = createMmapFile(dataDir + "/couplingStep")
    mmTimeList = createMmapFile(dataDir + "/timeList")
    #streamを定義
    streamToParent = pyCoupling.stream(mmCommFile, commonSize,
                                       waitCount, waitTime )
    #引数を保存
    beginning(couplingDict)
    #終了？
    if (float(timeName) >= float(couplingDict["endTime"])):
        exit()
    #計算開始時？（初期化の有無）
    if (float(beforeTime) == float(startTime) 
        and float(startTime) == 0.0):
        #初回起動時の設定
        deleteMappingTriFile()
    #連続計算
    solidCaseDir = os.path.abspath(solidCaseDir)
    stat = setPress_getDisp()
    if stat != "":
        raise NameError(stat)
